"""

XYZ-aware commands

Just add the XYZGridCmdSet to the default character cmdset to override
the commands with XYZ-aware equivalents.

"""

from collections import namedtuple

from django.conf import settings

from evennia import CmdSet, InterruptCommand, default_cmds
from evennia.commands.default import building
from evennia.contrib.grid.xyzgrid.xyzgrid import get_xyzgrid
from evennia.contrib.grid.xyzgrid.xyzroom import XYZRoom
from evennia.utils import ansi
from evennia.utils.utils import class_from_module, delay, list_to_string

COMMAND_DEFAULT_CLASS = class_from_module(settings.COMMAND_DEFAULT_CLASS)


# temporary store of goto/path data when using the auto-stepper
PathData = namedtuple("PathData", ("target", "xymap", "directions", "step_sequence", "task"))


class CmdXYZTeleport(building.CmdTeleport):
    """
    teleport object to another location

    Usage:
      tel/switch [<object> to||=] <target location>
      tel/switch [<object> to||=] (X,Y[,Z])

    Examples:
      tel Limbo
      tel/quiet box = Limbo
      tel/tonone box
      tel (3, 3, the small cave)
      tel (4, 1)   # on the same map
      tel/map Z|mapname

    Switches:
      quiet  - don't echo leave/arrive messages to the source/target
               locations for the move.
      intoexit - if target is an exit, teleport INTO
                 the exit object instead of to its destination
      tonone - if set, teleport the object to a None-location. If this
               switch is set, <target location> is ignored.
               Note that the only way to retrieve
               an object from a None location is by direct #dbref
               reference. A puppeted object cannot be moved to None.
      loc - teleport object to the target's location instead of its contents
      map - show coordinate map of given Zcoord/mapname.

    Teleports an object somewhere. If no object is given, you yourself are
    teleported to the target location. If (X,Y) or (X,Y,Z) coordinates
    are given, the target is a location on the XYZGrid.

    """

    def _search_by_xyz(self, inp):
        inp = inp.strip("()")
        X, Y, *Z = inp.split(",", 2)
        if Z:
            # Z was specified
            Z = Z[0]
        else:
            # use current location's Z, if it exists
            try:
                xyz = self.caller.location.xyz
            except AttributeError:
                self.caller.msg(
                    "Z-coordinate is also required since you are not currently "
                    "in a room with a Z coordinate of its own."
                )
                raise InterruptCommand
            else:
                Z = xyz[2]
        # search by coordinate
        X, Y, Z = str(X).strip(), str(Y).strip(), str(Z).strip()
        try:
            self.destination = XYZRoom.objects.get_xyz(xyz=(X, Y, Z))
        except XYZRoom.DoesNotExist:
            self.caller.msg(f"Found no target XYZRoom at ({X},{Y},{Z}).")
            raise InterruptCommand

    def parse(self):
        default_cmds.MuxCommand.parse(self)
        self.obj_to_teleport = self.caller
        self.destination = None

        if self.rhs:
            self.obj_to_teleport = self.caller.search(self.lhs, global_search=True)
            if not self.obj_to_teleport:
                self.caller.msg("Did not find object to teleport.")
                raise InterruptCommand
            if all(char in self.rhs for char in ("(", ")", ",")):
                # search by (X,Y) or (X,Y,Z)
                self._search_by_xyz(self.rhs)
            else:
                # fallback to regular search by name/alias
                self.destination = self.caller.search(self.rhs, global_search=True)

        elif self.lhs:
            if all(char in self.lhs for char in ("(", ")", ",")):
                self._search_by_xyz(self.lhs)
            else:
                self.destination = self.caller.search(self.lhs, global_search=True)


class CmdXYZOpen(building.CmdOpen):
    """
    open a new exit from the current room

    Usage:
      open <new exit>[;alias;..][:typeclass] [,<return exit>[;alias;..][:typeclass]]] = <destination>
      open <new exit>[;alias;..][:typeclass] [,<return exit>[;alias;..][:typeclass]]] = (X,Y,Z)

    Handles the creation of exits. If a destination is given, the exit
    will point there. The destination can also be given as an (X,Y,Z) coordinate on the
    XYZGrid - this command is used to link non-grid rooms to the grid and vice-versa.

    The <return exit> argument sets up an exit at the destination leading back to the current room.
    Apart from (X,Y,Z) coordinate, destination name can be given both as a #dbref and a name, if
    that name is globally unique.

    Examples:
        open kitchen = Kitchen
        open north, south = Town Center
        open cave mouth;cave = (3, 4, the small cave)

    """

    def parse(self):
        building.ObjManipCommand.parse(self)

        self.location = self.caller.location
        if not self.args or not self.rhs:
            self.caller.msg(
                "Usage: open <new exit>[;alias...][:typeclass]"
                "[,<return exit>[;alias..][:typeclass]]] "
                "= <destination or (X,Y,Z)>"
            )
            raise InterruptCommand
        if not self.location:
            self.caller.msg("You cannot create an exit from a None-location.")
            raise InterruptCommand

        if all(char in self.rhs for char in ("(", ")", ",")):
            # search by (X,Y) or (X,Y,Z)
            inp = self.rhs.strip("()")
            X, Y, *Z = inp.split(",", 2)
            if not Z:
                self.caller.msg("A full (X,Y,Z) coordinate must be given for the destination.")
                raise InterruptCommand
            Z = Z[0]
            # search by coordinate
            X, Y, Z = str(X).strip(), str(Y).strip(), str(Z).strip()
            try:
                self.destination = XYZRoom.objects.get_xyz(xyz=(X, Y, Z))
            except XYZRoom.DoesNotExist:
                self.caller.msg(f"Found no target XYZRoom at ({X},{Y},{Z}).")
                raise InterruptCommand
        else:
            # regular search query
            self.destination = self.caller.search(self.rhs, global_search=True)
            if not self.destination:
                raise InterruptCommand

        self.exit_name = self.lhs_objs[0]["name"]
        self.exit_aliases = self.lhs_objs[0]["aliases"]
        self.exit_typeclass = self.lhs_objs[0]["option"]


class CmdGoto(COMMAND_DEFAULT_CLASS):
    """
    Go to a named location in this area via the shortest path.

    Usage:
        path <location>      - find shortest path to target location (don't move)
        goto <location>      - auto-move to target location, using shortest path
        path                 - show current target location and shortest path
        goto                 - abort current goto, otherwise show current path
        path clear           - clear current path

    Finds the shortest route to a location in your current area and
    can then automatically walk you there.

    Builders can optionally specify a specific grid coordinate (X,Y) to go to.

    """

    key = "goto"
    aliases = "path"
    help_category = "General"
    locks = "cmd:all()"

    # how quickly to step (seconds)
    auto_step_delay = 2
    default_xyz_path_interrupt_msg = "Pathfinding interrupted here."

    def _search_by_xyz(self, inp, xyz_start):
        inp = inp.strip("()")
        X, Y = inp.split(",", 2)
        Z = xyz_start[2]
        # search by coordinate
        X, Y, Z = str(X).strip(), str(Y).strip(), str(Z).strip()
        try:
            return XYZRoom.objects.get_xyz(xyz=(X, Y, Z))
        except XYZRoom.DoesNotExist:
            self.caller.msg(f"Could not find a room at ({X},{Y}) (Z={Z}).")
            return None

    def _search_by_key_and_alias(self, inp, xyz_start):
        Z = xyz_start[2]
        candidates = list(XYZRoom.objects.filter_xyz(xyz=("*", "*", Z)))
        return self.caller.search(inp, candidates=candidates)

    def _auto_step(
        self,
        caller,
        session,
        target=None,
        xymap=None,
        directions=None,
        step_sequence=None,
        step=True,
    ):
        path_data = caller.ndb.xy_path_data

        if target:
            # start/replace an old path if we provide the data for it
            if path_data and path_data.task and path_data.task.active():
                # stop any old task in its tracks
                path_data.task.cancel()
            path_data = caller.ndb.xy_path_data = PathData(
                target=target,
                xymap=xymap,
                directions=directions,
                step_sequence=step_sequence,
                task=None,
            )

        if step and path_data:
            step_sequence = path_data.step_sequence

            try:
                direction = path_data.directions.pop(0)
                current_node = path_data.step_sequence.pop(0)
                first_link = path_data.step_sequence.pop(0)
            except IndexError:
                caller.msg("Target reached.", session=session)
                caller.ndb.xy_path_data = None
                return

            # verfy our current location against the expected location
            expected_xyz = (current_node.X, current_node.Y, current_node.Z)
            location = caller.location
            try:
                xyz_start = location.xyz
            except AttributeError:
                caller.ndb.xy_path_data = None
                caller.msg("Goto aborted - outside of area.", session=session)
                return

            if xyz_start != expected_xyz:
                # we are not where we expected to be (maybe the user moved
                # manually)  - we must recalculate the path to target
                caller.msg("Path changed - recalculating ('goto' to abort)", session=session)

                try:
                    xyz_end = path_data.target.xyz
                except AttributeError:
                    caller.ndb.xy_path_data = None
                    caller.msg("Goto aborted - target outside of area.", session=session)
                    return

                if xyz_start[2] != xyz_end[2]:
                    # can't go to another map
                    caller.ndb.xy_path_data = None
                    caller.msg("Goto aborted - target outside of area.", session=session)
                    return

                # recalculate path
                xy_start = xyz_start[:2]
                xy_end = xyz_end[:2]
                directions, step_sequence = path_data.xymap.get_shortest_path(xy_start, xy_end)

                # try again with this path, rebuilding the data
                try:
                    direction = directions.pop(0)
                    current_node = step_sequence.pop(0)
                    first_link = step_sequence.pop(0)
                except IndexError:
                    caller.msg("Target reached.", session=session)
                    caller.ndb.xy_path_data = None
                    return

                path_data = caller.ndb.xy_path_data = PathData(
                    target=path_data.target,
                    xymap=path_data.xymap,
                    directions=directions,
                    step_sequence=step_sequence,
                    task=None,
                )
            # the map can itself tell the stepper to stop the auto-step prematurely
            interrupt_node_or_link = None

            # pop any extra links up until the next node - these are
            # not useful when dealing with exits
            while step_sequence:
                if not interrupt_node_or_link and step_sequence[0].interrupt_path:
                    interrupt_node_or_link = step_sequence[0]
                if hasattr(step_sequence[0], "node_index"):
                    break
                step_sequence.pop(0)

            # the exit name does not need to be the same as the cardinal direction!
            exit_name, *_ = first_link.spawn_aliases.get(
                direction, current_node.direction_spawn_defaults.get(direction, ("unknown",))
            )

            exit_obj = caller.search(exit_name)
            if not exit_obj:
                # extra safety measure to avoid trying to walk over and over
                # if there's something wrong with the exit's name
                caller.msg(f"No exit '{exit_name}' found at current location. Aborting goto.")
                caller.ndb.xy_path_data = None
                return

            if interrupt_node_or_link:
                # premature stop of pathfind-step because of map node/link of interrupt type
                if hasattr(interrupt_node_or_link, "node_index"):
                    message = exit_obj.destination.attributes.get(
                        "xyz_path_interrupt_msg", default=self.default_xyz_path_interrupt_msg
                    )
                    # we move into the node/room and then stop
                    caller.execute_cmd(exit_name, session=session)
                else:
                    # if the link is interrupted we don't cross it at all
                    message = exit_obj.attributes.get(
                        "xyz_path_interrupt_msg", default=self.default_xyz_path_interrupt_msg
                    )
                caller.msg(message)
                return

            # do the actual move - we use the command to allow for more obvious overrides
            caller.execute_cmd(exit_name, session=session)

            # namedtuples are unmutables, so we recreate and store
            # with the new task
            caller.ndb.xy_path_data = PathData(
                target=path_data.target,
                xymap=path_data.xymap,
                directions=path_data.directions,
                step_sequence=path_data.step_sequence,
                task=delay(self.auto_step_delay, self._auto_step, caller, session),
            )

    def func(self):
        """
        Implement command
        """

        caller = self.caller
        goto_mode = self.cmdname == "goto"

        # check if we have an existing path
        path_data = caller.ndb.xy_path_data

        if not self.args:
            if path_data:
                target_name = path_data.target.get_display_name(caller)
                task = path_data.task
                if goto_mode:
                    if task and task.active():
                        task.cancel()
                        caller.msg(f"Aborted auto-walking to {target_name}.")
                        return
                # goto/path-command will show current path
                current_path = list_to_string([f"|w{step}|n" for step in path_data.directions])
                moving = "(moving)" if task and task.active() else ""
                caller.msg(f"Path to {target_name}{moving}: {current_path}")
            else:
                caller.msg("Usage: goto|path [<location>]")
            return

        if not goto_mode and self.args == "clear" and path_data:
            # in case there is a target location 'clear', this is only
            # used if path data already exists.
            caller.ndb.xy_path_data = None
            caller.msg("Cleared goto-path.")
            return

        # find target
        xyzgrid = get_xyzgrid()
        try:
            xyz_start = caller.location.xyz
        except AttributeError:
            self.caller.msg("Cannot path-find since the current location is not on the grid.")
            return

        allow_xyz_query = caller.locks.check_lockstring(caller, "perm(Builder)")
        if allow_xyz_query and all(char in self.args for char in ("(", ")", ",")):
            # search by (X,Y)
            target = self._search_by_xyz(self.args, xyz_start)
            if not target:
                return
        else:
            # search by normal key/alias
            target = self._search_by_key_and_alias(self.args, xyz_start)
            if not target:
                return
        try:
            xyz_end = target.xyz
        except AttributeError:
            self.caller.msg("Target location is not on the grid and cannot be auto-walked to.")
            return

        xymap = xyzgrid.get_map(xyz_start[2])
        # we only need the xy coords once we have the map
        xy_start = xyz_start[:2]
        xy_end = xyz_end[:2]
        directions, step_sequence = xymap.get_shortest_path(xy_start, xy_end)

        caller.msg(
            f"There are {len(directions)} steps to {target.get_display_name(caller)}: "
            f"|w{list_to_string(directions, endsep='|n, and finally|w')}|n"
        )

        # create data for display and start stepping if we used goto
        self._auto_step(
            caller,
            self.session,
            target=target,
            xymap=xymap,
            directions=directions,
            step_sequence=step_sequence,
            step=goto_mode,
        )


class CmdMap(COMMAND_DEFAULT_CLASS):
    """
    Show a map of an area

    Usage:
        map [Zcoord]
        map list

    This is a builder-command.

    """

    key = "map"
    locks = "cmd:perm(Builders)"

    def func(self):
        """Implement command"""

        xyzgrid = get_xyzgrid()
        Z = None

        if not self.args:
            # show current area's map
            location = self.caller.location
            try:
                xyz = location.xyz
            except AttributeError:
                self.caller.msg("Your current location is not on the grid.")
                return
            Z = xyz[2]

        elif self.args.strip().lower() == "list":
            xymaps = "\n ".join(str(repr(xymap)) for xymap in xyzgrid.all_maps())
            self.caller.msg(f"Maps (Z coords) on the grid:\n |w{xymaps}")
            return

        else:
            Z = self.args

        xymap = xyzgrid.get_map(Z)
        if not xymap:
            self.caller.msg(
                f"XYMap '{Z}' is not found on the grid. Try 'map list' to see "
                "available maps/Zcoords."
            )
            return

        self.caller.msg(ansi.raw(xymap.mapstring))


class XYZGridCmdSet(CmdSet):
    """
    Cmdset for easily adding the above cmds to the character cmdset.

    """

    key = "xyzgrid_cmdset"

    def at_cmdset_creation(self):
        self.add(CmdXYZTeleport())
        self.add(CmdXYZOpen())
        self.add(CmdGoto())
        self.add(CmdMap())


# Optional fly/dive commands to move between maps (enable
# full 3D-grid movements)


class CmdFlyAndDive(COMMAND_DEFAULT_CLASS):
    """
    Fly or Dive up and down.

    Usage:
      fly
      dive

    Will fly up one room or dive down one room at your current position. If
    there is no room above/below you, your movement will fail.

    """

    key = "fly or dive"
    aliases = ("fly", "dive")

    def func(self):
        caller = self.caller

        action = self.cmdname

        try:
            xyz_start = caller.location.xyz
        except AttributeError:
            caller.msg(f"You cannot {action} here.")
            return
        try:
            zcoord = int(xyz_start[2])
        except ValueError:
            caller.msg(f"You cannot {action} here.")
            return

        if action == "fly":
            diff = 1
            direction = "upwards"
            from_direction = "below"
            error_message = "Can't fly here - you'd hit your head."
        elif action == "dive":
            diff = -1
            direction = "downwards"
            from_direction = "above"
            error_message = "Can't dive here - you'd just fall flat on the ground."
        else:
            caller.msg("You must decide if you want to |wfly|n up or |wdive|n down.")
            return

        target_coord = (str(xyz_start[0]), str(xyz_start[1]), zcoord + diff)
        try:
            target = XYZRoom.objects.get_xyz(xyz=(target_coord))
        except XYZRoom.DoesNotExist:
            # no available room above/below to fly/dive to
            caller.msg(error_message)
            return
        # action succeeds, we have a target. One could picture being able to
        # lock certain rooms from flight/dive, here we allow it as long as there
        # is a suitable room above/below.
        caller.location.msg_contents(f"$You() {action} {direction}.", from_obj=caller)
        caller.move_to(target, quiet=True)
        target.msg_contents(
            f"$You() {action} from {from_direction}.", from_obj=caller, exclude=[caller]
        )


class XYZGridFlyDiveCmdSet(CmdSet):
    """
    Optional cmdset if you want the fly/dive commands to move in a 3D environment.

    """

    key = "xyzgrid_flydive_cmdset"

    def at_cmdset_creation(self):
        self.add(CmdFlyAndDive())



"""
Test of EvAdventure Rooms

"""

from evennia import DefaultExit, create_object
from evennia.utils.ansi import strip_ansi
from evennia.utils.test_resources import EvenniaTestCase

from ..characters import EvAdventureCharacter
from ..rooms import EvAdventureRoom


class EvAdventureRoomTest(EvenniaTestCase):
    def setUp(self):
        self.char = create_object(EvAdventureCharacter, key="TestChar")

    def test_map(self):
        center_room = create_object(EvAdventureRoom, key="room_center")
        n_room = create_object(EvAdventureRoom, key="room_n")
        create_object(DefaultExit, key="north", location=center_room, destination=n_room)
        ne_room = create_object(EvAdventureRoom, key="room_ne")
        create_object(DefaultExit, key="northeast", location=center_room, destination=ne_room)
        e_room = create_object(EvAdventureRoom, key="room_e")
        create_object(DefaultExit, key="east", location=center_room, destination=e_room)
        se_room = create_object(EvAdventureRoom, key="room_se")
        create_object(DefaultExit, key="southeast", location=center_room, destination=se_room)
        s_room = create_object(EvAdventureRoom, key="room_")
        create_object(DefaultExit, key="south", location=center_room, destination=s_room)
        sw_room = create_object(EvAdventureRoom, key="room_sw")
        create_object(DefaultExit, key="southwest", location=center_room, destination=sw_room)
        w_room = create_object(EvAdventureRoom, key="room_w")
        create_object(DefaultExit, key="west", location=center_room, destination=w_room)
        nw_room = create_object(EvAdventureRoom, key="room_nw")
        create_object(DefaultExit, key="northwest", location=center_room, destination=nw_room)

        desc = center_room.return_appearance(self.char)

        expected = r"""
  o o o
   \|/
  o-@-o
   /|\
  o o o
room_center
You see nothing special.
Exits: north, northeast, east, southeast, south, southwest, west, and northwest"""

        result = "\n".join(part.rstrip() for part in strip_ansi(desc).split("\n"))
        expected = "\n".join(part.rstrip() for part in expected.split("\n"))
        # print(result)
        # print(expected)

        self.assertEqual(result, expected)



"""
This is a little routine for viewing the sql queries that are executed by a given
query as well as count them for optimization testing.

"""

import os
import sys

# sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))
# os.environ["DJANGO_SETTINGS_MODULE"] = "game.settings"
from django.db import connection


def count_queries(exec_string, setup_string):
    """
    Display queries done by exec_string. Use setup_string
    to setup the environment to test.
    """

    exec(setup_string)

    num_queries_old = len(connection.queries)
    exec(exec_string)
    nqueries = len(connection.queries) - num_queries_old

    for query in connection.queries[-nqueries if nqueries else 1 :]:
        print(query["time"], query["sql"])
    print("Number of queries: %s" % nqueries)


if __name__ == "__main__":
    # setup tests here

    setup_string = """
from evennia.objects.models import ObjectDB
g = ObjectDB.objects.get(db_key="Griatch")
"""
    exec_string = """
g.tags.all()
"""
    count_queries(exec_string, setup_string)



"""
Character class.

"""

from evennia.objects.objects import DefaultCharacter
from evennia.typeclasses.attributes import AttributeProperty
from evennia.utils.evform import EvForm
from evennia.utils.evtable import EvTable
from evennia.utils.logger import log_trace
from evennia.utils.utils import lazy_property

from . import rules
from .equipment import EquipmentError, EquipmentHandler
from .quests import EvAdventureQuestHandler


class LivingMixin:
    """
    Mixin class to use for all living things.

    """

    is_pc = False

    @property
    def hurt_level(self):
        """
        String describing how hurt this character is.
        """
        percent = max(0, min(100, 100 * (self.hp / self.hp_max)))
        if 95 < percent <= 100:
            return "|gPerfect|n"
        elif 80 < percent <= 95:
            return "|gScraped|n"
        elif 60 < percent <= 80:
            return "|GBruised|n"
        elif 45 < percent <= 60:
            return "|yHurt|n"
        elif 30 < percent <= 45:
            return "|yWounded|n"
        elif 15 < percent <= 30:
            return "|rBadly wounded|n"
        elif 1 < percent <= 15:
            return "|rBarely hanging on|n"
        elif percent == 0:
            return "|RCollapsed!|n"

    def heal(self, hp, healer=None):
        """
        Heal by a certain amount of HP.

        """
        damage = self.hp_max - self.hp
        healed = min(damage, hp)
        self.hp += healed

        if healer is self:
            self.msg(f"|gYou heal yourself for {healed} health.|n")
        elif healer:
            self.msg(f"|g{healer.key} heals you for {healed} health.|n")
        else:
            self.msg(f"You are healed for {healed} health.")

    def at_attacked(self, attacker, **kwargs):
        """
        Called when being attacked / combat starts.

        """
        pass

    def at_damage(self, damage, attacker=None):
        """
        Called when attacked and taking damage.

        """
        self.hp -= damage

    def at_defeat(self):
        """
        Called when this living thing reaches HP 0.

        """
        # by default, defeat means death
        self.at_death()

    def at_death(self):
        """
        Called when this living thing dies.

        """
        pass

    def at_pay(self, amount):
        """
        Get coins, but no more than we actually have.

        """
        amount = min(amount, self.coins)
        self.coins -= amount
        return amount

    def at_looted(self, looter):
        """
        Called when being looted (after defeat).

        Args:
            looter (Object): The one doing the looting.

        """
        max_steal = rules.dice.roll("1d10")
        stolen = self.at_pay(max_steal)

        looter.coins += stolen

        self.location.msg_contents(
            f"$You(looter) loots $You() for {stolen} coins!",
            from_obj=self,
            mapping={"looter": looter},
        )

    def pre_loot(self, defeated_enemy):
        """
        Called just before looting an enemy.

        Args:
            defeated_enemy (Object): The enemy soon to loot.

        Returns:
            bool: If False, no looting is allowed.

        """
        pass

    def at_do_loot(self, defeated_enemy):
        """
        Called when looting another entity.

        Args:
            defeated_enemy: The thing to loot.

        """
        defeated_enemy.at_looted(self)

    def post_loot(self, defeated_enemy):
        """
        Called just after having looted an enemy.

        Args:
            defeated_enemy (Object): The enemy just looted.

        """
        pass


class EvAdventureCharacter(LivingMixin, DefaultCharacter):
    """
    A Character for use with EvAdventure.

    """

    is_pc = True

    # these are the ability bonuses. Defense is always 10 higher
    strength = AttributeProperty(default=1)
    dexterity = AttributeProperty(default=1)
    constitution = AttributeProperty(default=1)
    intelligence = AttributeProperty(default=1)
    wisdom = AttributeProperty(default=1)
    charisma = AttributeProperty(default=1)

    hp = AttributeProperty(default=4)
    hp_max = AttributeProperty(default=4)
    level = AttributeProperty(default=1)
    coins = AttributeProperty(default=0)  # copper coins

    xp = AttributeProperty(default=0)
    xp_per_level = 1000

    @lazy_property
    def equipment(self):
        """Allows to access equipment like char.equipment.worn"""
        return EquipmentHandler(self)

    @lazy_property
    def quests(self):
        """Access and track quests"""
        return EvAdventureQuestHandler(self)

    @property
    def weapon(self):
        return self.equipment.weapon

    @property
    def armor(self):
        return self.equipment.armor

    def at_pre_object_receive(self, moved_object, source_location, **kwargs):
        """
        Hook called by Evennia before moving an object here. Return False to abort move.

        Args:
            moved_object (Object): Object to move into this one (that is, into inventory).
            source_location (Object): Source location moved from.
            **kwargs: Passed from move operation; the `move_type` is useful; if someone is giving
                us something (`move_type=='give'`) we want to ask first.

        Returns:
            bool: If move should be allowed or not.

        """
        # this will raise EquipmentError if inventory is full
        return self.equipment.validate_slot_usage(moved_object)

    def at_object_receive(self, moved_object, source_location, **kwargs):
        """
        Hook called by Evennia as an object is moved here. We make sure it's added
        to the equipment handler.

        Args:
            moved_object (Object): Object to move into this one (that is, into inventory).
            source_location (Object): Source location moved from.
            **kwargs: Passed from move operation; unused here.

        """
        try:
            self.equipment.add(moved_object)
        except EquipmentError as err:
            log_trace(f"at_object_receive error: {err}")

    def at_pre_object_leave(self, leaving_object, destination, **kwargs):
        """
        Hook called when dropping an item. We don't allow to drop wielded/worn items
        (need to unwield/remove them first). Return False to

        """
        return True

    def at_object_leave(self, moved_object, destination, **kwargs):
        """
        Called just before an object leaves from inside this object

        Args:
            moved_obj (Object): The object leaving
            destination (Object): Where `moved_obj` is going.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        """
        self.equipment.remove(moved_object)

    def at_defeat(self):
        """
        This happens when character drops <= 0 HP. For Characters, this means rolling on
        the death table.

        """
        if self.location.allow_death:
            rules.dice.roll_death(self)
        else:
            self.location.msg_contents("|y$You() $conj(yield), beaten and out of the fight.|n")
            self.hp = self.hp_max

    def at_death(self):
        """
        Called when character dies.

        """
        self.location.msg_contents(
            "|r$You() $conj(collapse) in a heap.\nDeath embraces you ...|n",
            from_obj=self,
        )

    def at_pre_loot(self):
        """
        Called before allowing to loot. Return False to block enemy looting.
        """
        # don't allow looting in pvp
        return not self.location.allow_pvp

    def at_looted(self, looter):
        """
        Called when being looted.

        """
        pass

    def add_xp(self, xp):
        """
        Add new XP.

        Args:
            xp (int): The amount of gained XP.

        Returns:
            bool: If a new level was reached or not.

        Notes:
            level 1 -> 2 = 1000 XP
            level 2 -> 3 = 2000 XP etc

        """
        self.xp += xp
        next_level_xp = self.level * self.xp_per_level
        return self.xp >= next_level_xp

    def level_up(self, *abilities):
        """
        Perform the level-up action.

        Args:
            *abilities (str): A set of abilities (like 'strength', 'dexterity' (normally 3)
                to upgrade by 1. Max is usually +10.
        Notes:
            We block increases above a certain value, but we don't raise an error here, that
            will need to be done earlier, when the user selects the ability to increase.

        """

        self.level += 1
        for ability in set(abilities[:3]):
            # limit to max amount allowed, each one unique
            try:
                # set at most to the max bonus
                current_bonus = getattr(self, ability)
                setattr(
                    self,
                    ability,
                    min(10, current_bonus + 1),
                )
            except AttributeError:
                pass

        # update hp
        self.hp_max = max(self.max_hp + 1, rules.dice.roll(f"{self.level}d8"))


# character sheet visualization


_SHEET = """
 +----------------------------------------------------------------------------+
 | Name: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx1xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx |
 +----------------------------------------------------------------------------+
 | STR: x2xxxxx  DEX: x3xxxxx  CON: x4xxxxx  WIS: x5xxxxx  CHA: x6xxxxx       |
 +----------------------------------------------------------------------------+
 | HP: x7xxxxx                                      XP: x8xxxxx  Level: x9x   |
 +----------------------------------------------------------------------------+
 | Desc: xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx |
 | xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxBxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx |
 | xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx |
 +----------------------------------------------------------------------------+
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccc1ccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 | cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc |
 +----------------------------------------------------------------------------+
    """


def get_character_sheet(character):
    """
    Generate a character sheet. This is grouped in a class in order to make
    it easier to override the look of the sheet.

    """

    @staticmethod
    def get(character):
        """
        Generate a character sheet from the character's stats.

        """
        equipment = character.equipment.all()
        # divide into chunks of max 10 length (to go into two columns)
        equipment_table = EvTable(
            table=[equipment[i : i + 10] for i in range(0, len(equipment), 10)]
        )
        form = EvForm({"FORMCHAR": "x", "TABLECHAR": "c", "SHEET": _SHEET})
        form.map(
            cells={
                1: character.key,
                2: f"+{character.strength}({character.strength + 10})",
                3: f"+{character.dexterity}({character.dexterity + 10})",
                4: f"+{character.constitution}({character.constitution + 10})",
                5: f"+{character.wisdom}({character.wisdom + 10})",
                6: f"+{character.charisma}({character.charisma + 10})",
                7: f"{character.hp}/{character.hp_max}",
                8: character.xp,
                9: character.level,
                "A": character.db.desc,
            },
            tables={
                1: equipment_table,
            },
        )
        return str(form)



from evennia import DefaultCharacter
from evennia.utils import inherits_from
from evennia.utils.test_resources import BaseEvenniaCommandTest

from . import character_creator


class TestCharacterCreator(BaseEvenniaCommandTest):
    def setUp(self):
        super().setUp()
        self.account.swap_typeclass(character_creator.ContribChargenAccount)
        self.account.unpuppet_all()

    def test_account_look(self):
        self.account.characters.add(self.char1)
        self.char1.db.chargen_step = "start"

        # check that correct output is returning
        output = self.account.at_look(target=self.account.characters.all(), session=self.session)
        # check that char1 is recognized as in progress
        self.assertIn("in progress", output)

    def test_char_create(self):
        with self.settings(START_LOCATION=f"#{self.room1.id}"):
            self.call(
                character_creator.ContribCmdCharCreate(),
                "",
                caller=self.account,
            )
        # verify menu was initialized
        menu = self.session.ndb._menutree
        self.assertNotEqual(menu, None)
        # verify character was created
        new_char = self.session.new_char
        self.assertTrue(inherits_from(new_char, DefaultCharacter))
        # verify character's "start location" was set
        self.assertEqual(self.session.new_char.db.prelogout_location, self.room1)

        # exit the menu, verify it resumes
        menu.parse_input("q")
        del self.session.new_char
        self.assertEqual(self.session.ndb._menutree, None)
        self.call(
            character_creator.ContribCmdCharCreate(),
            "",
            caller=self.account,
        )
        # should be the same new char
        self.assertEqual(new_char, self.session.new_char)



"""
Unix-like Command style parent

Evennia contribution, Vincent Le Geoff 2017

This module contains a command class that allows for unix-style command syntax in-game, using
--options, positional arguments and stuff like -n 10 etc similarly to a unix command. It might not
the best syntax for the average player but can be really useful for builders when they need to have
a single command do many things with many options. It uses the ArgumentParser from Python's standard
library under the hood.

To use, inherit `UnixCommand` from this module from your own commands. You need
to override two methods:

- The `init_parser` method, which adds options to the parser. Note that you should normally
    *not* override the normal `parse` method when inheriting from `UnixCommand`.
- The `func` method, called to execute the command once parsed (like any Command).

Here's a short example:

```python
from evennia.contrib.base_systems.unixcommand import UnixCommand


class CmdPlant(UnixCommand):

    '''
    Plant a tree or plant.

    This command is used to plant something in the room you are in.

    Examples:
      plant orange -a 8
      plant strawberry --hidden
      plant potato --hidden --age 5

    '''

    key = "plant"

    def init_parser(self):
        "Add the arguments to the parser."
        # 'self.parser' inherits `argparse.ArgumentParser`
        self.parser.add_argument("key",
                help="the key of the plant to be planted here")
        self.parser.add_argument("-a", "--age", type=int,
                default=1, help="the age of the plant to be planted")
        self.parser.add_argument("--hidden", action="store_true",
                help="should the newly-planted plant be hidden to players?")

    def func(self):
        "func is called only if the parser succeeded."
        # 'self.opts' contains the parsed options
        key = self.opts.key
        age = self.opts.age
        hidden = self.opts.hidden
        self.msg("Going to plant '{}', age={}, hidden={}.".format(
                key, age, hidden))
```

To see the full power of argparse and the types of supported options, visit
[the documentation of argparse](https://docs.python.org/2/library/argparse.html).

"""

import argparse
import shlex
from textwrap import dedent

from evennia import Command, InterruptCommand
from evennia.utils.ansi import raw


class ParseError(Exception):

    """An error occurred during parsing."""

    pass


class UnixCommandParser(argparse.ArgumentParser):

    """A modifier command parser for unix commands.

    This parser is used to replace `argparse.ArgumentParser`.  It
    is aware of the command calling it, and can more easily report to
    the caller.  Some features (like the "brutal exit" of the original
    parser) are disabled or replaced.  This parser is used by UnixCommand
    and creating one directly isn't recommended nor necessary.  Even
    adding a sub-command will use this replaced parser automatically.

    """

    def __init__(self, prog, description="", epilog="", command=None, **kwargs):
        """
        Build a UnixCommandParser with a link to the command using it.

        Args:
            prog (str): the program name (usually the command key).
            description (str): a very brief line to show in the usage text.
            epilog (str): the epilog to show below options.
            command (Command): the command calling the parser.

        Keyword Args:
            Additional keyword arguments are directly sent to
            `argparse.ArgumentParser`.  You will find them on the
            [parser's documentation](https://docs.python.org/2/library/argparse.html).

        Note:
            It's doubtful you would need to create this parser manually.
            The `UnixCommand` does that automatically.  If you create
            sub-commands, this class will be used.

        """
        prog = prog or command.key
        super().__init__(
            prog=prog, description=description, conflict_handler="resolve", add_help=False, **kwargs
        )
        self.command = command
        self.post_help = epilog

        def n_exit(code=None, msg=None):
            raise ParseError(msg)

        self.exit = n_exit

        # Replace the -h/--help
        self.add_argument(
            "-h", "--help", nargs=0, action=HelpAction, help="display the command help"
        )

    def format_usage(self):
        """Return the usage line.

        Note:
            This method is present to return the raw-escaped usage line,
            in order to avoid unintentional color codes.

        """
        return raw(super().format_usage())

    def format_help(self):
        """Return the parser help, including its epilog.

        Note:
            This method is present to return the raw-escaped help,
            in order to avoid unintentional color codes.  Color codes
            in the epilog (the command docstring) are supported.

        """
        autohelp = raw(super().format_help())
        return "\n" + autohelp + "\n" + self.post_help

    def print_usage(self, file=None):
        """Print the usage to the caller.

        Args:
            file (file-object): not used here, the caller is used.

        Note:
            This method will override `argparse.ArgumentParser`'s in order
            to not display the help on stdout or stderr, but to the
            command's caller.

        """
        if self.command:
            self.command.msg(self.format_usage().strip())

    def print_help(self, file=None):
        """Print the help to the caller.

        Args:
            file (file-object): not used here, the caller is used.

        Note:
            This method will override `argparse.ArgumentParser`'s in order
            to not display the help on stdout or stderr, but to the
            command's caller.

        """
        if self.command:
            self.command.msg(self.format_help().strip())


class HelpAction(argparse.Action):

    """Override the -h/--help action in the default parser.

    Using the default -h/--help will call the exit function in different
    ways, preventing the entire help message to be provided.  Hence
    this override.

    """

    def __call__(self, parser, namespace, values, option_string=None):
        """If asked for help, display to the caller."""
        if parser.command:
            parser.command.msg(parser.format_help().strip())
            parser.exit(0, "")


class UnixCommand(Command):
    """
    Unix-type commands, supporting short and long options.

    This command syntax uses the Unix-style commands with short options
    (-X) and long options (--something).  The `argparse` module is
    used to parse the command.

    In order to use it, you should override two methods:
    - `init_parser`: this method is called when the command is created.
      It can be used to set options in the parser.  `self.parser`
      contains the `argparse.ArgumentParser`, so you can add arguments
      here.
    - `func`: this method is called to execute the command, but after
      the parser has checked the arguments given to it are valid.
      You can access the namespace of valid arguments in `self.opts`
      at this point.

    The help of UnixCommands is derived from the docstring, in a
    slightly different way than usual: the first line of the docstring
    is used to represent the program description (the very short
    line at the top of the help message).  The other lines below are
    used as the program's "epilog", displayed below the options.  It
    means in your docstring, you don't have to write the options.
    They will be automatically provided by the parser and displayed
    accordingly.  The `argparse` module provides a default '-h' or
    '--help' option on the command.  Typing |whelp commandname|n will
    display the same as |wcommandname -h|n, though this behavior can
    be changed.

    """

    def __init__(self, **kwargs):
        """
        The lockhandler works the same as for objects.
        optional kwargs will be set as properties on the Command at runtime,
        overloading evential same-named class properties.

        """
        super().__init__(**kwargs)

        # Create the empty UnixCommandParser, inheriting argparse.ArgumentParser
        lines = dedent(self.__doc__.strip("\n")).splitlines()
        description = lines[0].strip()
        epilog = "\n".join(lines[1:]).strip()
        self.parser = UnixCommandParser(None, description, epilog, command=self)

        # Fill the argument parser
        self.init_parser()

    def init_parser(self):
        """
        Configure the argument parser, adding in options.

        Note:
            This method is to be overridden in order to add options
            to the argument parser.  Use `self.parser`, which contains
            the `argparse.ArgumentParser`.  You can, for instance,
            use its `add_argument` method.

        """
        pass

    def func(self):
        """Override to handle the command execution."""
        pass

    def get_help(self, caller, cmdset):
        """
        Return the help message for this command and this caller.

        Args:
            caller (Object or Player): the caller asking for help on the command.
            cmdset (CmdSet): the command set (if you need additional commands).

        Returns:
            docstring (str): the help text to provide the caller for this command.

        """
        return self.parser.format_help()

    def parse(self):
        """
        Process arguments provided in `self.args`.

        Note:
            You should not override this method.  Consider overriding
            `init_parser` instead.

        """
        try:
            self.opts = self.parser.parse_args(shlex.split(self.args))
        except ParseError as err:
            msg = str(err)
            if msg:
                self.msg(msg)
            raise InterruptCommand



"""

Admin commands

"""

import re
import time

from django.conf import settings

import evennia
from evennia.server.models import ServerConfig
from evennia.utils import class_from_module, evtable, logger, search

COMMAND_DEFAULT_CLASS = class_from_module(settings.COMMAND_DEFAULT_CLASS)

PERMISSION_HIERARCHY = [p.lower() for p in settings.PERMISSION_HIERARCHY]

# limit members for API inclusion
__all__ = (
    "CmdBoot",
    "CmdBan",
    "CmdUnban",
    "CmdEmit",
    "CmdNewPassword",
    "CmdPerm",
    "CmdWall",
    "CmdForce",
)


class CmdBoot(COMMAND_DEFAULT_CLASS):
    """
    kick an account from the server.

    Usage
      boot[/switches] <account obj> [: reason]

    Switches:
      quiet - Silently boot without informing account
      sid - boot by session id instead of name or dbref

    Boot an account object from the server. If a reason is
    supplied it will be echoed to the user unless /quiet is set.
    """

    key = "boot"
    switch_options = ("quiet", "sid")
    locks = "cmd:perm(boot) or perm(Admin)"
    help_category = "Admin"

    def func(self):
        """Implementing the function"""
        caller = self.caller
        args = self.args

        if not args:
            caller.msg("Usage: boot[/switches] <account> [:reason]")
            return

        if ":" in args:
            args, reason = [a.strip() for a in args.split(":", 1)]
        else:
            args, reason = args, ""

        boot_list = []

        if "sid" in self.switches:
            # Boot a particular session id.
            sessions = evennia.SESSION_HANDLER.get_sessions(True)
            for sess in sessions:
                # Find the session with the matching session id.
                if sess.sessid == int(args):
                    boot_list.append(sess)
                    break
        else:
            # Boot by account object
            pobj = search.account_search(args)
            if not pobj:
                caller.msg(f"Account {args} was not found.")
                return
            pobj = pobj[0]
            if not pobj.access(caller, "boot"):
                caller.msg(f"You don't have the permission to boot {pobj.key}.")
                return
            # we have a bootable object with a connected user
            matches = evennia.SESSION_HANDLER.sessions_from_account(pobj)
            for match in matches:
                boot_list.append(match)

        if not boot_list:
            caller.msg("No matching sessions found. The Account does not seem to be online.")
            return

        # Carry out the booting of the sessions in the boot list.

        feedback = None
        if "quiet" not in self.switches:
            feedback = f"You have been disconnected by {caller.name}.\n"
            if reason:
                feedback += f"\nReason given: {reason}"

        for session in boot_list:
            session.msg(feedback)
            session.account.disconnect_session_from_account(session)

        if pobj and boot_list:
            logger.log_sec(
                f"Booted: {pobj} (Reason: {reason}, Caller: {caller}, IP: {self.session.address})."
            )


# regex matching IP addresses with wildcards, eg. 233.122.4.*
IPREGEX = re.compile(r"[0-9*]{1,3}\.[0-9*]{1,3}\.[0-9*]{1,3}\.[0-9*]{1,3}")


def list_bans(cmd, banlist):
    """
    Helper function to display a list of active bans. Input argument
    is the banlist read into the two commands ban and unban below.

    Args:
        cmd (Command): Instance of the Ban command.
        banlist (list): List of bans to list.
    """
    if not banlist:
        return "No active bans were found."

    table = cmd.styled_table("|wid", "|wname/ip", "|wdate", "|wreason")
    for inum, ban in enumerate(banlist):
        table.add_row(str(inum + 1), ban[0] and ban[0] or ban[1], ban[3], ban[4])
    return f"|wActive bans:|n\n{table}"


class CmdBan(COMMAND_DEFAULT_CLASS):
    """
    ban an account from the server

    Usage:
      ban [<name or ip> [: reason]]

    Without any arguments, shows numbered list of active bans.

    This command bans a user from accessing the game. Supply an optional
    reason to be able to later remember why the ban was put in place.

    It is often preferable to ban an account from the server than to
    delete an account with accounts/delete. If banned by name, that account
    account can no longer be logged into.

    IP (Internet Protocol) address banning allows blocking all access
    from a specific address or subnet. Use an asterisk (*) as a
    wildcard.

    Examples:
      ban thomas             - ban account 'thomas'
      ban/ip 134.233.2.111   - ban specific ip address
      ban/ip 134.233.2.*     - ban all in a subnet
      ban/ip 134.233.*.*     - even wider ban

    A single IP filter can be easy to circumvent by changing computers
    or requesting a new IP address. Setting a wide IP block filter with
    wildcards might be tempting, but remember that it may also
    accidentally block innocent users connecting from the same country
    or region.

    """

    key = "ban"
    aliases = ["bans"]
    locks = "cmd:perm(ban) or perm(Developer)"
    help_category = "Admin"

    def func(self):
        """
        Bans are stored in a serverconf db object as a list of
        dictionaries:
          [ (name, ip, ipregex, date, reason),
            (name, ip, ipregex, date, reason),...  ]
        where name and ip are set by the user and are shown in
        lists. ipregex is a converted form of ip where the * is
        replaced by an appropriate regex pattern for fast
        matching. date is the time stamp the ban was instigated and
        'reason' is any optional info given to the command. Unset
        values in each tuple is set to the empty string.
        """
        banlist = ServerConfig.objects.conf("server_bans")
        if not banlist:
            banlist = []

        if not self.args or (
            self.switches and not any(switch in ("ip", "name") for switch in self.switches)
        ):
            self.msg(list_bans(self, banlist))
            return

        now = time.ctime()
        reason = ""
        if ":" in self.args:
            ban, reason = self.args.rsplit(":", 1)
        else:
            ban = self.args
        ban = ban.lower()
        ipban = IPREGEX.findall(ban)
        if not ipban:
            # store as name
            typ = "Name"
            bantup = (ban, "", "", now, reason)
        else:
            # an ip address.
            typ = "IP"
            ban = ipban[0]
            # replace * with regex form and compile it
            ipregex = ban.replace(".", "\.")
            ipregex = ipregex.replace("*", "[0-9]{1,3}")
            ipregex = re.compile(r"%s" % ipregex)
            bantup = ("", ban, ipregex, now, reason)

        ret = yield (f"Are you sure you want to {typ}-ban '|w{ban}|n' [Y]/N?")
        if str(ret).lower() in ("no", "n"):
            self.msg("Aborted.")
            return

        # save updated banlist
        banlist.append(bantup)
        ServerConfig.objects.conf("server_bans", banlist)
        self.msg(f"{typ}-ban '|w{ban}|n' was added. Use |wunban|n to reinstate.")
        logger.log_sec(
            f"Banned {typ}: {ban.strip()} (Caller: {self.caller}, IP: {self.session.address})."
        )


class CmdUnban(COMMAND_DEFAULT_CLASS):
    """
    remove a ban from an account

    Usage:
      unban <banid>

    This will clear an account name/ip ban previously set with the ban
    command.  Use this command without an argument to view a numbered
    list of bans. Use the numbers in this list to select which one to
    unban.

    """

    key = "unban"
    locks = "cmd:perm(unban) or perm(Developer)"
    help_category = "Admin"

    def func(self):
        """Implement unbanning"""

        banlist = ServerConfig.objects.conf("server_bans")

        if not self.args:
            self.msg(list_bans(self, banlist))
            return

        try:
            num = int(self.args)
        except Exception:
            self.msg("You must supply a valid ban id to clear.")
            return

        if not banlist:
            self.msg("There are no bans to clear.")
        elif not (0 < num < len(banlist) + 1):
            self.msg(f"Ban id |w{self.args}|n was not found.")
        else:
            # all is ok, ask, then clear ban
            ban = banlist[num - 1]
            value = (" ".join([s for s in ban[:2]])).strip()

            ret = yield (f"Are you sure you want to unban {num}: '|w{value}|n' [Y]/N?")
            if str(ret).lower() in ("n", "no"):
                self.msg("Aborted.")
                return

            del banlist[num - 1]
            ServerConfig.objects.conf("server_bans", banlist)
            self.msg(f"Cleared ban {num}: '{value}'")
            logger.log_sec(
                f"Unbanned: {value.strip()} (Caller: {self.caller}, IP: {self.session.address})."
            )


class CmdEmit(COMMAND_DEFAULT_CLASS):
    """
    admin command for emitting message to multiple objects

    Usage:
      emit[/switches] [<obj>, <obj>, ... =] <message>
      remit           [<obj>, <obj>, ... =] <message>
      pemit           [<obj>, <obj>, ... =] <message>

    Switches:
      room     -  limit emits to rooms only (default)
      accounts -  limit emits to accounts only
      contents -  send to the contents of matched objects too

    Emits a message to the selected objects or to
    your immediate surroundings. If the object is a room,
    send to its contents. remit and pemit are just
    limited forms of emit, for sending to rooms and
    to accounts respectively.
    """

    key = "emit"
    aliases = ["pemit", "remit"]
    switch_options = ("room", "accounts", "contents")
    locks = "cmd:perm(emit) or perm(Builder)"
    help_category = "Admin"

    def func(self):
        """Implement the command"""

        caller = self.caller
        args = self.args

        if not args:
            string = "Usage: "
            string += "\nemit[/switches] [<obj>, <obj>, ... =] <message>"
            string += "\nremit           [<obj>, <obj>, ... =] <message>"
            string += "\npemit           [<obj>, <obj>, ... =] <message>"
            caller.msg(string)
            return

        rooms_only = "rooms" in self.switches
        accounts_only = "accounts" in self.switches
        send_to_contents = "contents" in self.switches

        # we check which command was used to force the switches
        if self.cmdstring == "remit":
            rooms_only = True
            send_to_contents = True
        elif self.cmdstring == "pemit":
            accounts_only = True

        if not self.rhs:
            message = self.args
            objnames = [caller.location.key]
        else:
            message = self.rhs
            objnames = self.lhslist

        # send to all objects
        for objname in objnames:
            obj = caller.search(objname, global_search=True)
            if not obj:
                return
            if rooms_only and obj.location is not None:
                caller.msg(f"{objname} is not a room. Ignored.")
                continue
            if accounts_only and not obj.has_account:
                caller.msg(f"{objname} has no active account. Ignored.")
                continue
            if obj.access(caller, "tell"):
                obj.msg(message)
                if send_to_contents and hasattr(obj, "msg_contents"):
                    obj.msg_contents(message)
                    caller.msg(f"Emitted to {objname} and contents:\n{message}")
                else:
                    caller.msg(f"Emitted to {objname}:\n{message}")
            else:
                caller.msg(f"You are not allowed to emit to {objname}.")


class CmdNewPassword(COMMAND_DEFAULT_CLASS):
    """
    change the password of an account

    Usage:
      userpassword <user obj> = <new password>

    Set an account's password.
    """

    key = "userpassword"
    locks = "cmd:perm(newpassword) or perm(Admin)"
    help_category = "Admin"

    def func(self):
        """Implement the function."""

        caller = self.caller

        if not self.rhs:
            self.msg("Usage: userpassword <user obj> = <new password>")
            return

        # the account search also matches 'me' etc.
        account = caller.search_account(self.lhs)
        if not account:
            return

        newpass = self.rhs

        # Validate password
        validated, error = account.validate_password(newpass)
        if not validated:
            errors = [e for suberror in error.messages for e in error.messages]
            string = "\n".join(errors)
            caller.msg(string)
            return

        account.set_password(newpass)
        account.save()
        self.msg(f"{account.name} - new password set to '{newpass}'.")
        if account.character != caller:
            account.msg(f"{caller.name} has changed your password to '{newpass}'.")
        logger.log_sec(
            f"Password Changed: {account} (Caller: {caller}, IP: {self.session.address})."
        )


class CmdPerm(COMMAND_DEFAULT_CLASS):
    """
    set the permissions of an account/object

    Usage:
      perm[/switch] <object> [= <permission>[,<permission>,...]]
      perm[/switch] *<account> [= <permission>[,<permission>,...]]

    Switches:
      del     -  delete the given permission from <object> or <account>.
      account -  set permission on an account (same as adding * to name)

    This command sets/clears individual permission strings on an object
    or account. If no permission is given, list all permissions on <object>.
    """

    key = "perm"
    aliases = "setperm"
    switch_options = ("del", "account")
    locks = "cmd:perm(perm) or perm(Developer)"
    help_category = "Admin"

    def func(self):
        """Implement function"""

        caller = self.caller
        switches = self.switches
        lhs, rhs = self.lhs, self.rhs

        if not self.args:
            string = "Usage: perm[/switch] object [ = permission, permission, ...]"
            caller.msg(string)
            return

        accountmode = "account" in self.switches or lhs.startswith("*")
        lhs = lhs.lstrip("*")

        if accountmode:
            obj = caller.search_account(lhs)
        else:
            obj = caller.search(lhs, global_search=True)
        if not obj:
            return

        if not rhs:
            if not obj.access(caller, "examine"):
                caller.msg("You are not allowed to examine this object.")
                return

            string = f"Permissions on |w{obj.key}|n: "
            if not obj.permissions.all():
                string += "<None>"
            else:
                string += ", ".join(obj.permissions.all())
                if (
                    hasattr(obj, "account")
                    and hasattr(obj.account, "is_superuser")
                    and obj.account.is_superuser
                ):
                    string += "\n(... but this object is currently controlled by a SUPERUSER! "
                    string += "All access checks are passed automatically.)"
            caller.msg(string)
            return

        # we supplied an argument on the form obj = perm
        locktype = "edit" if accountmode else "control"
        if not obj.access(caller, locktype):
            accountstr = "account" if accountmode else "object"
            caller.msg(f"You are not allowed to edit this {accountstr}'s permissions.")
            return

        caller_result = []
        target_result = []
        if "del" in switches:
            # delete the given permission(s) from object.
            for perm in self.rhslist:
                obj.permissions.remove(perm)
                if obj.permissions.get(perm):
                    caller_result.append(
                        f"\nPermissions {perm} could not be removed from {obj.name}."
                    )
                else:
                    caller_result.append(
                        f"\nPermission {perm} removed from {obj.name} (if they existed)."
                    )
                    target_result.append(
                        f"\n{caller.name} revokes the permission(s) {perm} from you."
                    )
                    logger.log_sec(
                        f"Permissions Deleted: {perm}, {obj} (Caller: {caller}, IP: {self.session.address})."
                    )
        else:
            # add a new permission
            permissions = obj.permissions.all()

            for perm in self.rhslist:
                # don't allow to set a permission higher in the hierarchy than
                # the one the caller has (to prevent self-escalation)
                if perm.lower() in PERMISSION_HIERARCHY and not obj.locks.check_lockstring(
                    caller, f"dummy:perm({perm})"
                ):
                    caller.msg(
                        "You cannot assign a permission higher than the one you have yourself."
                    )
                    return

                if perm in permissions:
                    caller_result.append(f"\nPermission '{perm}' is already defined on {obj.name}.")
                else:
                    obj.permissions.add(perm)
                    plystring = "the Account" if accountmode else "the Object/Character"
                    caller_result.append(
                        f"\nPermission '{perm}' given to {obj.name} ({plystring})."
                    )
                    target_result.append(
                        f"\n{caller.name} gives you ({obj.name}, {plystring}) the permission '{perm}'."
                    )
                    logger.log_sec(
                        f"Permissions Added: {perm}, {obj} (Caller: {caller}, IP: {self.session.address})."
                    )

        caller.msg("".join(caller_result).strip())
        if target_result:
            obj.msg("".join(target_result).strip())


class CmdWall(COMMAND_DEFAULT_CLASS):
    """
    make an announcement to all

    Usage:
      wall <message>

    Announces a message to all connected sessions
    including all currently unlogged in.
    """

    key = "wall"
    locks = "cmd:perm(wall) or perm(Admin)"
    help_category = "Admin"

    def func(self):
        """Implements command"""
        if not self.args:
            self.msg("Usage: wall <message>")
            return
        message = f'{self.caller.name} shouts "{self.args}"'
        self.msg("Announcing to all connected sessions ...")
        evennia.SESSION_HANDLER.announce_all(message)


class CmdForce(COMMAND_DEFAULT_CLASS):
    """
    forces an object to execute a command

    Usage:
        force <object>=<command string>

    Example:
        force bob=get stick
    """

    key = "force"
    locks = "cmd:perm(spawn) or perm(Builder)"
    help_category = "Building"
    perm_used = "edit"

    def func(self):
        """Implements the force command"""
        if not self.lhs or not self.rhs:
            self.msg("You must provide a target and a command string to execute.")
            return
        targ = self.caller.search(self.lhs)
        if not targ:
            return
        if not targ.access(self.caller, self.perm_used):
            self.msg(f"You don't have permission to force {targ} to execute commands.")
            return
        targ.execute_cmd(self.rhs)
        self.msg(f"You have forced {targ} to: {self.rhs}")



# -*- coding: utf-8 -*-

"""
Testing various individual functionalities in the server package.

"""
import unittest

from django.test import TestCase
from django.test.runner import DiscoverRunner

from evennia.server.throttle import Throttle
from evennia.utils.test_resources import BaseEvenniaTest

from ..deprecations import check_errors


class MockSettings(object):
    """
    Class for simulating django.conf.settings. Created with a single value, and then sets the required
    WEBSERVER_ENABLED setting to True or False depending if we're testing WEBSERVER_PORTS.
    """

    def __init__(self, setting, value=None):
        setattr(self, setting, value)
        if setting == "WEBSERVER_PORTS":
            self.WEBSERVER_ENABLED = True
        else:
            self.WEBSERVER_ENABLED = False


class TestDeprecations(TestCase):
    """
    Class for testing deprecations.check_errors.
    """

    deprecated_settings = (
        "CMDSET_DEFAULT",
        "CMDSET_OOC",
        "BASE_COMM_TYPECLASS",
        "COMM_TYPECLASS_PATHS",
        "CHARACTER_DEFAULT_HOME",
        "OBJECT_TYPECLASS_PATHS",
        "SCRIPT_TYPECLASS_PATHS",
        "ACCOUNT_TYPECLASS_PATHS",
        "CHANNEL_TYPECLASS_PATHS",
        "SEARCH_MULTIMATCH_SEPARATOR",
        "TIME_SEC_PER_MIN",
        "TIME_MIN_PER_HOUR",
        "TIME_HOUR_PER_DAY",
        "TIME_DAY_PER_WEEK",
        "TIME_WEEK_PER_MONTH",
        "TIME_MONTH_PER_YEAR",
        "GAME_DIRECTORY_LISTING",
    )

    def test_check_errors(self):
        """
        All settings in deprecated_settings should raise a DeprecationWarning if they exist.
        WEBSERVER_PORTS raises an error if the iterable value passed does not have a tuple as its
        first element.
        """
        for setting in self.deprecated_settings:
            self.assertRaises(DeprecationWarning, check_errors, MockSettings(setting))
        # test check for WEBSERVER_PORTS having correct value
        self.assertRaises(
            DeprecationWarning, check_errors, MockSettings("WEBSERVER_PORTS", value=["not a tuple"])
        )


class ThrottleTest(BaseEvenniaTest):
    """
    Class for testing the connection/IP throttle.
    """

    def test_throttle(self):
        ips = ("256.256.256.257", "257.257.257.257", "258.258.258.258")
        kwargs = {"name": "testing", "limit": 5, "timeout": 15 * 60}

        throttle = Throttle(**kwargs)

        for ip in ips:
            # Throttle should not be engaged by default
            self.assertFalse(throttle.check(ip))

            # Pretend to fail a bunch of events
            for x in range(50):
                obj = throttle.update(ip)
                self.assertFalse(obj)

            # Next ones should be blocked
            self.assertTrue(throttle.check(ip))

            for x in range(throttle.cache_size * 2):
                obj = throttle.update(ip)
                self.assertFalse(obj)

            # Should still be blocked
            self.assertTrue(throttle.check(ip))

            # Number of values should be limited by cache size
            self.assertEqual(throttle.cache_size, len(throttle.get(ip)))

        cache = throttle.get()

        # Make sure there are entries for each IP
        self.assertEqual(len(ips), len(cache.keys()))

        # There should only be (cache_size * num_ips) total in the Throttle cache
        self.assertEqual(sum([len(cache[x]) for x in cache.keys()]), throttle.cache_size * len(ips))

        # Make sure the cache is populated
        self.assertTrue(throttle.get())

        # Remove the test IPs from the throttle cache
        # (in case persistent storage was configured by the user)
        for ip in ips:
            self.assertTrue(throttle.remove(ip))

        # Make sure the cache is empty
        self.assertFalse(throttle.get())



"""
EvAdventure Shop system.


A shop is run by an NPC. It can provide one or more of several possible services:

- Buy from a pre-set list of (possibly randomized) items. Cost is based on the item's value,
  adjusted by how stingy the shopkeeper is. When bought this way, the item is
  generated on the fly and passed to the player character's inventory. Inventory files are
  a list of prototypes, normally from a prototype-file. A random selection of items from each
  inventory file is available.
- Sell items to the shop for a certain percent of their value. One could imagine being able
  to buy back items again, but we will instead _destroy_ sold items, so as to remove them
  from circulation. In-game we can say it's because the merchants collect the best stuff
  to sell to collectors in the big city later. Each merchant buys a certain subset of items
  based on their tags.
- Buy a service. For a cost, a certain action is performed for the character; this applies
  immediately when bought. The most notable services are healing and converting coin to XP.
- Buy rumors - this is echoed to the player for a price. Different merchants could have
  different rumors (or randomized ones).
- Quest - gain or hand in a quest for a merchant.

All shops are menu-driven. One starts talking to the npc and will then end up in their shop
interface.


This is a series of menu nodes meant to be added as a mapping via
`EvAdventureShopKeeper.create(menudata={},...)`.

To make this pluggable, the shopkeeper start page will analyze the available nodes
and auto-add options to all nodes in the three named `node_start_*`. The last part of the
node name will be the name of the option capitalized, with underscores replaced by spaces, so
`node_start_sell_items` will become a top-level option `Sell items`.



"""

from dataclasses import dataclass

from evennia.prototypes.prototypes import search_prototype
from evennia.prototypes.spawner import flatten_prototype, spawn
from evennia.utils.evmenu import list_node
from evennia.utils.logger import log_err, log_trace

from .enums import ObjType, WieldLocation
from .equipment import EquipmentError

# ------------------------------------ Buying from an NPC


@dataclass
class BuyItem:
    """
    Storage container for storing generic info about an item for sale. This means it can be used
    both for real objects and for prototypes without constantly having to track which is which.

    """

    # skipping typehints here since we are not using them anywhere else

    # available for all buyable items
    key = ""
    desc = ""
    obj_type = ObjType.GEAR
    size = 1
    value = 0
    use_slot = WieldLocation.BACKPACK

    uses = None
    quality = None
    attack_type = None
    defense_type = None
    damage_roll = None

    # references the original (always only one of the two)
    obj = None
    prototype = None

    @staticmethod
    def create_from_obj(obj, shopkeeper):
        """
        Build a new BuyItem container from a real db obj.

        Args:
            obj (EvAdventureObject): An object to analyze.
            shopkeeper (EvAdventureShopKeeper): The shopkeeper.

        Returns:
            BuyItem: A general representation of the original data.

        """
        try:
            # mandatory
            key = obj.key
            desc = obj.db.desc
            obj_type = obj.obj_type
            size = obj.size
            use_slot = obj.use_slot
            value = obj.value * shopkeeper.upsell_factor
        except AttributeError:
            # not a buyable item
            log_trace("Not a buyable item")
            return None

        # getting optional properties

        return BuyItem(
            key=key,
            desc=desc,
            obj_type=obj_type,
            size=size,
            use_slot=use_slot,
            value=value,
            # optional fields
            uses=getattr(obj, "uses", None),
            quality=getattr(obj, "quality", None),
            attack_type=getattr(obj, "attack_type", None),
            defense_type=getattr(obj, "defense_type", None),
            damage_roll=getattr(obj, "damage_roll", None),
            # back-reference (don't set prototype)
            obj=obj,
        )

    @staticmethod
    def create_from_prototype(self, prototype_or_key, shopkeeper):
        """
        Build a new BuyItem container from a prototype.

        Args:
            prototype (dict or key): An Evennia prototype dict or the key of one
                registered with the system. This is assumed to be a full prototype,
                including having parsed and included parentage.

        Returns:
            BuyItem: A general representation of the original data.

        """

        def _get_attr_value(key, prot, optional=True):
            """
            We want the attribute's value, which is always in the `attrs` field of
            the prototype.

            """
            attr = [tup for tup in prot.get("attrs", ()) if tup[0] == key]
            try:
                return attr[0][1]
            except IndexError:
                if optional:
                    return None
                raise

        if isinstance(prototype_or_key, dict):
            prototype = prototype_or_key
        else:
            # make sure to generate a 'full' prototype with all inheritance applied ('flattened'),
            # otherwise we will not get inherited data when we analyze it.
            prototype = flatten_prototype(search_prototype(key=prototype_or_key))

        if not prototype:
            log_err(f"No valid prototype '{prototype_or_key}' found")
            return None

        try:
            # at this point we should have a full, flattened prototype ready to spawn. It must
            # contain all fields needed for buying
            key = prototype["key"]
            desc = _get_attr_value("desc", prototype, optional=False)
            obj_type = _get_attr_value("obj_type", prototype, optional=False)
            size = _get_attr_value("size", prototype, optional=False)
            use_slot = _get_attr_value("use_slot", prototype, optional=False)
            value = int(
                _get_attr_value("value", prototype, optional=False) * shopkeeper.upsell_factor
            )
        except (KeyError, IndexError):
            # not a buyable item
            log_trace("Not a buyable item")
            return None

        return BuyItem(
            key=key,
            desc=desc,
            obj_type=obj_type,
            size=size,
            use_slot=use_slot,
            value=value,
            # optional fields
            uses=_get_attr_value("uses", prototype),
            quality=_get_attr_value("quality", prototype),
            attack_type=_get_attr_value("attack_type", prototype),
            defense_type=_get_attr_value("defense_type", prototype),
            damage_roll=_get_attr_value("damage_roll", prototype),
            # back-reference (don't set obj)
            prototype=prototype,
        )

    def __str__(self):
        """
        Get the short description to show in buy list.

        """
        return f"{self.key} [|y{self.value}|n coins]"

    def get_detail(self):
        """
        Get more info when looking at the item.

        """
        return f"""
|c{self.key}|n  Cost: |y{self.value}|n coins

{self.desc}

Slots: |w{self.size}|n Used from: |w{self.use_slot.value}|n
Quality: |w{self.quality}|n Uses: |wself.uses|n
Attacks using: |w{self.attack_type.value}|n against |w{self.defense_type.value}|n
Damage roll: |w{self.damage_roll}"""

    def to_obj(self):
        """
        Convert this into an actual database object that we can trade. This either means
        using the stored `.prototype` to spawn a new instance of the object, or to
        use the `.obj` reference to get the already existing object.

        """
        if self.obj:
            return self.obj
        return spawn(self.prototype)


def _get_or_create_buymap(caller, shopkeep):
    """
    Helper that fetches or creates the mapping of `{"short description": BuyItem, ...}`
    we need for the buy menu. We cache it on the `_evmenu` object on the caller.

    """
    if not caller.ndb._evmenu.buymap:
        # buymap not in cache - build it and store in memory on _evmenu object - this way
        # it will be removed automatically when the menu closes. We will need to reset this
        # when the shopkeep buys new things.
        # items carried by the shopkeep are sellable (these are items already created, such as
        # things sold to the shopkeep earlier). We
        obj_wares = [BuyItem.create_from_obj(obj) for obj in list(shopkeep.contents)]
        prototype_wares = [
            BuyItem.create_from_prototype(prototype)
            for prototype in shopkeep.common_ware_prototypes
        ]
        wares = obj_wares + prototype_wares
        caller.ndb._evmenu.buymap = {str(ware): ware for ware in wares if ware}

    return caller.ndb._evmenu.buymap


# Helper functions for building the shop listings and select a ware to buy
def _get_all_wares_to_buy(caller, raw_string, **kwargs):
    """
    This helper is used by `EvMenu.list_node` to build the list of items to buy.

    We rely on `**kwargs` being forwarded from `node_start_buy`, which in turns contains
    the `npc` kwarg pointing to the shopkeeper (`caller` is the one doing the buying).

    """
    shopkeep = kwargs["npc"]
    buymap = _get_or_create_buymap(caller, shopkeep)
    return [ware_desc for ware_desc in buymap]


def _select_ware_to_buy(caller, selected_ware_desc, **kwargs):
    """
    This helper is used by `EvMenu.list_node` to operate on what the user selected.
    We return `item` in the kwargs to the `node_select_buy` node.

    """
    shopkeep = kwargs["npc"]
    buymap = _get_or_create_buymap(caller, shopkeep)
    kwargs["item"] = buymap[selected_ware_desc]

    return "node_confirm_buy", kwargs


def _back_to_previous_node(caller, raw_string, **kwargs):
    """
    Back to previous node is achieved by returning a node of None.

    """
    return None, kwargs


def _buy_ware(caller, raw_string, **kwargs):
    """
    Complete the purchase of a ware. At this point the money is deducted
    and the item is either spawned from a prototype or simply moved from
    the sellers inventory to that of the buyer.

    We will have kwargs `item` and `npc` passed along to refer to the BuyItem we bought
    and the shopkeep selling it.

    """
    item = kwargs["item"]  # a BuyItem instance
    shopkeep = kwargs["npc"]

    # exchange money
    caller.coins -= item.value
    shopkeep += item.value

    # get the item - if not enough room, dump it on the ground
    obj = item.to_obj()
    try:
        caller.equipment.add(obj)
    except EquipmentError as err:
        obj.location = caller.location
        caller.msg(err)
        caller.msg(f"|w{obj.key} ends up on the ground.|n")

    caller.msg("|gYou bought |w{obj.key}|g for |y{item.value}|g coins.|n")


@list_node(_get_all_wares_to_buy, select=_select_ware_to_buy, pagesize=40)
def node_start_buy(caller, raw_string, **kwargs):
    """
    Menu node for the caller to buy items from the shopkeep. This assumes `**kwargs` contains
    a kwarg `npc` referencing the npc/shopkeep being talked to.

    Items available to sell are a combination of items in the shopkeep's inventory and
    the list of `prototypes` stored in the Shopkeep's "common_ware_prototypes` Attribute. In the
    latter case, the properties will be extracted from the prototype when inspecting it (object will
    only spawn when bought).

    """
    coins = caller.coins
    used_slots = caller.equipment.count_slots()
    max_slots = caller.equipment.max_slots

    text = (
        f'"Seeing something you like?" [you have |y{coins}|n coins, '
        f"using |b{used_slots}/{max_slots}|n slots]"
    )
    # this will be in addition to the options generated by the list-node
    extra_options = [{"key": ("[c]ancel", "b", "c", "cancel"), "goto": "node_start"}]

    return text, extra_options


def node_confirm_buy(caller, raw_string, **kwargs):
    """
    Menu node reached when a user selects an item in the buy menu. The `item` passed
    along in `**kwargs` is the selected item (see `_select_ware_to_buy`, where this is injected).

    """
    # this was injected in _select_ware_to_buy. This is an BuyItem instance.
    item = kwargs["item"]

    coins = caller.coins
    used_slots = caller.equipment.count_slots()
    max_slots = caller.equipment.max_slots

    text = item.get_detail()
    text += f"\n\n[You have |y{coins}|n coins] and are using |b{used_slots}/{max_slots}|n slots]"

    options = []

    if caller.coins >= item.value and item.size <= (max_slots - used_slots):
        options.append({"desc": f"Buy [{item.value} coins]", "goto": (_buy_ware, kwargs)})
    options.append({"desc": "Cancel", "goto": (_back_to_previous_node, kwargs)})

    return text, options


# node tree to inject for buying things
node_tree_buy = {"node_start_buy": node_start_buy, "node_confirm_buy": node_confirm_buy}


# ------------------------------------------------- Selling to an NPC


def _get_or_create_sellmap(self, caller, shopkeep):
    if not caller.ndb._evmenu.sellmap:
        # no sellmap, build one anew

        sellmap = {}
        for obj, wieldlocation in caller.equipment.all():
            key = obj.key
            value = int(obj.value * shopkeep.miser_factor)
            if value > 0 and obj.obj_type is not ObjType.QUEST:
                sellmap[f"|w{key}|n [{wieldlocation.value}] - sell price |y{value}|n coins"] = (
                    obj,
                    value,
                )
        caller.ndb._evmenu.sellmap = sellmap

    sellmap = caller.ndb._evmenu.sellmap

    return sellmap


def _get_all_wares_to_sell(caller, raw_string, **kwargs):
    """
    Get all wares available to sell from caller's inventory. We need to build a
    mapping between the descriptors and the items.

    """
    shopkeep = kwargs["npc"]
    sellmap = _get_or_create_sellmap(caller, shopkeep)
    return [ware_desc for ware_desc in sellmap]


def _sell_ware(caller, raw_string, **kwargs):
    """
    Complete the sale of a ware. This is were money is gained and the item is removed.

    We will have kwargs `item`, `value` and `npc` passed along to refer to the inventory item we
    sold, its (adjusted) sales cost and the shopkeep buying it.

    """
    item = kwargs["item"]
    value = kwargs["value"]
    shopkeep = kwargs["npc"]

    # move item to shopkeep
    obj = caller.equipment.remove(item)
    obj.location = shopkeep

    # exchange money - shopkeep always have money to pay, so we don't deduct from them
    caller.coins += value

    caller.msg("|gYou sold |w{obj.key}|g for |y{value}|g coins.|n")


def _select_ware_to_sell(caller, selected_ware_desc, **kwargs):
    """
    Selected one ware to sell. Figure out which one it is using the sellmap.
    Store the result as "item" kwarg.

    """
    shopkeep = kwargs["npc"]
    sellmap = _get_or_create_sellmap(caller, shopkeep)
    kwargs["item"], kwargs["value"] = sellmap[selected_ware_desc]

    return "node_examine_sell", kwargs


@list_node(_get_all_wares_to_sell, select=_select_ware_to_sell, pagesize=20)
def node_start_sell(caller, raw_string, **kwargs):
    """
    The start-level node for selling items from the user's inventory. This assumes
    `**kwargs` contains a kwarg `npc` referencing the npc/shopkeep being talked to.

    Items available to sell are all items in the player's equipment handler, including
    things in their hands.

    """
    coins = caller.coins
    used_slots = caller.equipment.count_slots()
    max_slots = caller.equipment.max_slots

    text = (
        f'"Anything you want to sell?" [you have |y{coins}|n coins, '
        f"using |b{used_slots}/{max_slots}|n slots]"
    )
    # this will be in addition to the options generated by the list-node
    extra_options = [{"key": ("[c]ancel", "b", "c", "cancel"), "goto": "node_start"}]

    return text, extra_options


def node_confirm_sell(caller, raw_string, **kwargs):
    """
    In this node we confirm the sell by first investigating the item we are about to sell.

    We have `item` and `value` available in kwargs here, added by `_select_ware_to_sell` earler.

    """
    item = kwargs["item"]
    value = kwargs["value"]

    coins = caller.coins
    used_slots = caller.equipment.count_slots()
    max_slots = caller.equipment.max_slots

    text = caller.equipment.get_obj_stats(item)
    text += f"\n\n[You have |y{coins}|n coins] and are using |b{used_slots}/{max_slots}|n slots]"

    options = (
        {"desc": f"Sell [{value} coins]", "goto": (_sell_ware, kwargs)},
        {"desc": "Cancel", "goto": (_back_to_previous_node, kwargs)},
    )

    return text, options


# node tree to inject for selling things
node_tree_sell = {"node_start_sell": node_start_sell, "node_confirm_sell": node_confirm_sell}


# Full shopkeep node tree - inject into ShopKeep NPC menu to add buy/sell submenus
node_tree_shopkeep = {**node_tree_buy, **node_tree_sell}



"""
LLM (Large Language Model) client, for communicating with an LLM backend. This can be used
for generating texts for AI npcs, or for fine-tuning the LLM on a given prompt.

Note that running a LLM locally requires a lot of power, and ideally a powerful GPU. Testing
this with CPU mode on a beefy laptop, still takes some 4s just on a very small model.

The server defaults to output suitable for a local server
https://github.com/oobabooga/text-generation-webui, but could be used for other LLM servers too.

See the LLM instructions on that page for how to set up the server. You'll also need
a model file - there are thousands to try out on https://huggingface.co/models (you want Text
Generation models specifically).

# Optional Evennia settings (if not given, these defaults are used)

DEFAULT_LLM_HOST = "http://localhost:5000"
DEFAULT_LLM_PATH = "/api/v1/generate"
DEFAULT_LLM_HEADERS = {"Content-Type": "application/json"}
DEFAULT_LLM_PROMPT_KEYNAME = "prompt"
DEFAULT_LLM_REQUEST_BODY = {...}   # see below, this controls how to prompt the LLM server.

"""

import json

from django.conf import settings
from twisted.internet import defer, protocol, reactor
from twisted.internet.defer import inlineCallbacks
from twisted.web.client import Agent, HTTPConnectionPool, _HTTP11ClientFactory
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer
from zope.interface import implementer

from evennia import logger
from evennia.utils.utils import make_iter

DEFAULT_LLM_HOST = "http://127.0.0.1:5000"
DEFAULT_LLM_PATH = "/api/v1/generate"
DEFAULT_LLM_HEADERS = {"Content-Type": "application/json"}
DEFAULT_LLM_PROMPT_KEYNAME = "prompt"
DEFAULT_LLM_API_TYPE = ""  # or openai
DEFAULT_LLM_REQUEST_BODY = {
    "max_new_tokens": 250,  # max number of tokens to generate
    "temperature": 0.7,  # higher = more random, lower = more predictable
}


@implementer(IBodyProducer)
class StringProducer:
    """
    Used for feeding a request body to the HTTP client.
    """

    def __init__(self, body):
        self.body = bytes(body, "utf-8")
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return defer.succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


class SimpleResponseReceiver(protocol.Protocol):
    """
    Used for pulling the response body out of an HTTP response.
    """

    def __init__(self, status_code, d):
        self.status_code = status_code
        self.buf = b""
        self.d = d

    def dataReceived(self, data):
        self.buf += data

    def connectionLost(self, reason=protocol.connectionDone):
        self.d.callback((self.status_code, self.buf))


class QuietHTTP11ClientFactory(_HTTP11ClientFactory):
    """
    Silences the obnoxious factory start/stop messages in the default client.
    """

    noisy = False


class LLMClient:
    """
    A client for communicating with an LLM server.

    """

    def __init__(self, on_bad_request=None):
        self._conn_pool = HTTPConnectionPool(reactor)
        self._conn_pool._factory = QuietHTTP11ClientFactory

        self.prompt_keyname = getattr(settings, "LLM_PROMPT_KEYNAME", DEFAULT_LLM_PROMPT_KEYNAME)
        self.hostname = getattr(settings, "LLM_HOST", DEFAULT_LLM_HOST)
        self.pathname = getattr(settings, "LLM_PATH", DEFAULT_LLM_PATH)
        self.headers = getattr(settings, "LLM_HEADERS", DEFAULT_LLM_HEADERS)
        self.request_body = getattr(settings, "LLM_REQUEST_BODY", DEFAULT_LLM_REQUEST_BODY)

        self.api_type = getattr(settings, "LLM_API_TYPE", DEFAULT_LLM_API_TYPE)

        self.agent = Agent(reactor, pool=self._conn_pool)

    def _format_request_body(self, prompt):
        """Structure the request body for the LLM server"""
        request_body = self.request_body.copy()

        prompt = "\n".join(make_iter(prompt))

        request_body[self.prompt_keyname] = prompt

        return request_body

    def _handle_llm_response_body(self, response):
        """Get the response body from the response"""
        d = defer.Deferred()
        response.deliverBody(SimpleResponseReceiver(response.code, d))
        return d

    def _handle_llm_error(self, failure):
        """Correctly handle server connection errors"""
        failure.trap(Exception)
        return (500, failure.getErrorMessage())

    def _get_response_from_llm_server(self, prompt):
        """Call the LLM server and handle the response/failure"""
        request_body = self._format_request_body(prompt)

        if settings.DEBUG:
            logger.log_info(f"LLM request body: {request_body}")

        d = self.agent.request(
            b"POST",
            bytes(self.hostname + self.pathname, "utf-8"),
            headers=Headers(self.headers),
            bodyProducer=StringProducer(json.dumps(request_body)),
        )

        d.addCallbacks(self._handle_llm_response_body, self._handle_llm_error)
        return d

    @inlineCallbacks
    def get_response(self, prompt):
        """
        Get a response from the LLM server for the given npc.

        Args:
            prompt (str or list): The prompt to send to the LLM server. If a list,
                this is assumed to be the chat history so far, and will be added to the
                prompt in a way suitable for the api.

        Returns:
            str: The generated text response. Will return an empty string
                if there is an issue with the server, in which case the
                the caller is expected to handle this gracefully.

        """
        status_code, response = yield self._get_response_from_llm_server(prompt)
        if status_code == 200:
            if settings.DEBUG:
                logger.log_info(f"LLM response: {response}")
            return json.loads(response)["results"][0]["text"]
        else:
            logger.log_err(f"LLM API error (status {status_code}): {response}")
            return ""



"""
EvAdventure character generation.

"""
from django.conf import settings

from evennia.objects.models import ObjectDB
from evennia.prototypes.spawner import spawn
from evennia.utils.create import create_object
from evennia.utils.evmenu import EvMenu

from .characters import EvAdventureCharacter
from .random_tables import chargen_tables
from .rules import dice

_ABILITIES = {
    "STR": "strength",
    "DEX": "dexterity",
    "CON": "constitution",
    "INT": "intelligence",
    "WIS": "wisdom",
    "CHA": "charisma",
}

_TEMP_SHEET = """
{name}

STR +{strength}
DEX +{dexterity}
CON +{constitution}
INT +{intelligence}
WIS +{wisdom}
CHA +{charisma}

{description}

Your belongings:
{equipment}
"""


class TemporaryCharacterSheet:
    """
    This collects all the rules for generating a new character. An instance of this class is used
    to pass around the current character state during character generation and also applied to
    the character at the end. This class instance can also be saved on the menu to make sure a user
    is not losing their half-created character.

    Note:
        In standard Knave, the character's attribute bonus is rolled randomly and will give a
        value 1-6; and there is no guarantee for 'equal' starting characters.

        Knave uses a d8 roll to get the initial hit points. We will follow the recommendation
        from the rule that we will use a minimum of 5 HP.

        We *will* roll random start equipment though. Contrary to standard Knave, we'll also
        randomly assign the starting weapon among a small selection of equal-dmg weapons (since
        there is no GM to adjudicate a different choice).

    """

    def _random_ability(self):
        return min(dice.roll("1d6"), dice.roll("1d6"), dice.roll("1d6"))

    def __init__(self):
        # name will likely be modified later
        self.name = dice.roll_random_table("1d282", chargen_tables["name"])

        # base attribute values
        self.strength = self._random_ability()
        self.dexterity = self._random_ability()
        self.constitution = self._random_ability()
        self.intelligence = self._random_ability()
        self.wisdom = self._random_ability()
        self.charisma = self._random_ability()

        # physical attributes (only for rp purposes)
        physique = dice.roll_random_table("1d20", chargen_tables["physique"])
        face = dice.roll_random_table("1d20", chargen_tables["face"])
        skin = dice.roll_random_table("1d20", chargen_tables["skin"])
        hair = dice.roll_random_table("1d20", chargen_tables["hair"])
        clothing = dice.roll_random_table("1d20", chargen_tables["clothing"])
        speech = dice.roll_random_table("1d20", chargen_tables["speech"])
        virtue = dice.roll_random_table("1d20", chargen_tables["virtue"])
        vice = dice.roll_random_table("1d20", chargen_tables["vice"])
        background = dice.roll_random_table("1d20", chargen_tables["background"])
        misfortune = dice.roll_random_table("1d20", chargen_tables["misfortune"])
        alignment = dice.roll_random_table("1d20", chargen_tables["alignment"])

        self.ability_changes = 0
        self.desc = (
            f"You are {physique} with a {face} face, {skin} skin, {hair} hair, {speech} speech, and"
            f" {clothing} clothing. You were a {background.title()}, but you were {misfortune} and"
            f" ended up a knave. You are {virtue} but also {vice}. You tend towards {alignment}."
        )

        # same for all
        self.hp_max = max(5, dice.roll("1d8"))
        self.hp = self.hp_max

        # random equipment
        self.armor = dice.roll_random_table("1d20", chargen_tables["armor"])

        _helmet_and_shield = dice.roll_random_table("1d20", chargen_tables["helmets and shields"])
        self.helmet = "helmet" if "helmet" in _helmet_and_shield else "none"
        self.shield = "shield" if "shield" in _helmet_and_shield else "none"

        self.weapon = dice.roll_random_table("1d20", chargen_tables["starting weapon"])

        self.backpack = [
            "ration",
            "ration",
            dice.roll_random_table("1d20", chargen_tables["dungeoning gear"]),
            dice.roll_random_table("1d20", chargen_tables["dungeoning gear"]),
            dice.roll_random_table("1d20", chargen_tables["general gear 1"]),
            dice.roll_random_table("1d20", chargen_tables["general gear 2"]),
        ]

    def show_sheet(self):
        """
        Show a temp character sheet, a compressed version of the real thing.

        """
        equipment = (
            str(item)
            for item in [self.armor, self.helmet, self.shield, self.weapon] + self.backpack
            if item
        )

        return _TEMP_SHEET.format(
            name=self.name,
            strength=self.strength,
            dexterity=self.dexterity,
            constitution=self.constitution,
            intelligence=self.intelligence,
            wisdom=self.wisdom,
            charisma=self.charisma,
            description=self.desc,
            equipment=", ".join(equipment),
        )

    def apply(self, account):
        """
        Once the chargen is complete, call this create and set up the character.

        """

        start_location = ObjectDB.objects.get_id(settings.START_LOCATION)
        default_home = ObjectDB.objects.get_id(settings.DEFAULT_HOME)
        permissions = settings.PERMISSION_ACCOUNT_DEFAULT
        # creating character with given abilities
        new_character = create_object(
            EvAdventureCharacter,
            key=self.name,
            location=start_location,
            home=default_home,
            permissions=permissions,
            attributes=(
                ("strength", self.strength),
                ("dexterity", self.dexterity),
                ("constitution", self.constitution),
                ("intelligence", self.intelligence),
                ("wisdom", self.wisdom),
                ("charisma", self.wisdom),
                ("hp", self.hp),
                ("hp_max", self.hp_max),
                ("desc", self.desc),
            ),
        )

        new_character.locks.add(
            "puppet:id(%i) or pid(%i) or perm(Developer) or pperm(Developer);delete:id(%i) or"
            " perm(Admin)" % (new_character.id, account.id, account.id)
        )
        # spawn equipment
        if self.weapon:
            weapon = spawn(self.weapon)
            new_character.equipment.move(weapon[0])
        if self.armor:
            armor = spawn(self.armor)
            new_character.equipment.move(armor[0])
        if self.shield:
            shield = spawn(self.shield)
            new_character.equipment.move(shield[0])
        if self.helmet:
            helmet = spawn(self.helmet)
            new_character.equipment.move(helmet[0])

        for item in self.backpack:
            item = spawn(item)
            new_character.equipment.move(item[0])

        return new_character


# chargen menu


def node_chargen(caller, raw_string, **kwargs):
    """
    This node is the central point of chargen. We return here to see our current
    sheet and break off to edit different parts of it.

    In Knave, not so much can be changed.
    """
    tmp_character = kwargs["tmp_character"]

    text = tmp_character.show_sheet()

    options = [{"desc": "Change your name", "goto": ("node_change_name", kwargs)}]
    if tmp_character.ability_changes <= 0:
        options.append(
            {
                "desc": "Swap two of your ability scores (once)",
                "goto": ("node_swap_abilities", kwargs),
            }
        )
    options.append(
        {"desc": "Accept and create character", "goto": ("node_apply_character", kwargs)},
    )

    return text, options


def _update_name(caller, raw_string, **kwargs):
    """
    Used by node_change_name below to check what user entered and update the name if appropriate.

    """
    if raw_string:
        tmp_character = kwargs["tmp_character"]
        tmp_character.name = raw_string.lower().capitalize()

    return "node_chargen", kwargs


def node_change_name(caller, raw_string, **kwargs):
    """
    Change the random name of the character.

    """
    tmp_character = kwargs["tmp_character"]

    text = (
        f"Your current name is |w{tmp_character.name}|n. Enter a new name or leave empty to abort."
    )

    options = {"key": "_default", "goto": (_update_name, kwargs)}

    return text, options


def _swap_abilities(caller, raw_string, **kwargs):
    """
    Used by node_swap_abilities to parse the user's input and swap ability
    values.

    """
    if raw_string:
        abi1, *abi2 = raw_string.split(" ", 1)
        if not abi2:
            caller.msg("That doesn't look right.")
            return None, kwargs
        abi2 = abi2[0]
        abi1, abi2 = abi1.upper().strip(), abi2.upper().strip()
        if abi1 not in _ABILITIES or abi2 not in _ABILITIES:
            caller.msg("Not a familiar set of abilites.")
            return None, kwargs

        # looks okay = swap values. We need to convert STR to strength etc
        tmp_character = kwargs["tmp_character"]
        abi1 = _ABILITIES[abi1]
        abi2 = _ABILITIES[abi2]
        abival1 = getattr(tmp_character, abi1)
        abival2 = getattr(tmp_character, abi2)

        setattr(tmp_character, abi1, abival2)
        setattr(tmp_character, abi2, abival1)

        tmp_character.ability_changes += 1

    return "node_chargen", kwargs


def node_swap_abilities(caller, raw_string, **kwargs):
    """
    One is allowed to swap the values of two abilities around, once.

    """
    tmp_character = kwargs["tmp_character"]

    text = f"""
Your current abilities:

STR +{tmp_character.strength}
DEX +{tmp_character.dexterity}
CON +{tmp_character.constitution}
INT +{tmp_character.intelligence}
WIS +{tmp_character.wisdom}
CHA +{tmp_character.charisma}

You can swap the values of two abilities around.
You can only do this once, so choose carefully!

To swap the values of e.g.  STR and INT, write |wSTR INT|n. Empty to abort.
"""

    options = {"key": "_default", "goto": (_swap_abilities, kwargs)}

    return text, options


def node_apply_character(caller, raw_string, **kwargs):
    """
    End chargen and create the character. We will also puppet it.

    """
    tmp_character = kwargs["tmp_character"]
    new_character = tmp_character.apply(caller)
    caller.characters.add(new_character)

    text = "Character created!"

    return text, None


def start_chargen(caller, session=None):
    """
    This is a start point for spinning up the chargen from a command later.

    """

    menutree = {
        "node_chargen": node_chargen,
        "node_change_name": node_change_name,
        "node_swap_abilities": node_swap_abilities,
        "node_apply_character": node_apply_character,
    }

    # this generates all random components of the character
    tmp_character = TemporaryCharacterSheet()

    EvMenu(
        caller,
        menutree,
        startnode="node_chargen",
        session=session,
        startnode_input=("", {"tmp_character": tmp_character}),
    )



#
# This sets up how models are displayed
# in the web admin interface.
#
from django import forms
from django.conf import settings
from django.contrib import admin, messages
from django.contrib.admin.utils import flatten_fieldsets
from django.contrib.admin.widgets import ForeignKeyRawIdWidget
from django.http import HttpResponseRedirect
from django.urls import path, reverse
from django.utils.html import format_html
from django.utils.translation import gettext as _

from evennia.accounts.models import AccountDB
from evennia.objects.models import ObjectDB

from . import utils as adminutils
from .attributes import AttributeInline
from .tags import TagInline


class ObjectAttributeInline(AttributeInline):
    """
    Defines inline descriptions of Attributes (experimental)

    """

    model = ObjectDB.db_attributes.through
    related_field = "objectdb"


class ObjectTagInline(TagInline):
    """
    Defines inline descriptions of Tags (experimental)

    """

    model = ObjectDB.db_tags.through
    related_field = "objectdb"


class ObjectCreateForm(forms.ModelForm):
    """
    This form details the look of the fields.

    """

    class Meta(object):
        model = ObjectDB
        fields = "__all__"

    db_key = forms.CharField(
        label="Name/Key",
        widget=forms.TextInput(attrs={"size": "78"}),
        help_text="Main identifier, like 'apple', 'strong guy', 'Elizabeth' etc. "
        "If creating a Character, check so the name is unique among characters!",
    )
    db_typeclass_path = forms.ChoiceField(
        label="Typeclass",
        initial={settings.BASE_OBJECT_TYPECLASS: settings.BASE_OBJECT_TYPECLASS},
        help_text="This is the Python-path to the class implementing the actual functionality. "
        f"<BR>If you are creating a Character you usually need <B>{settings.BASE_CHARACTER_TYPECLASS}</B> "
        "or a subclass of that. <BR>If your custom class is not found in the list, it may not be imported "
        "into Evennia yet.",
        choices=lambda: adminutils.get_and_load_typeclasses(parent=ObjectDB),
    )

    db_lock_storage = forms.CharField(
        label="Locks",
        required=False,
        widget=forms.Textarea(attrs={"cols": "100", "rows": "2"}),
        help_text="In-game lock definition string. If not given, defaults will be used. "
        "This string should be on the form "
        "<i>type:lockfunction(args);type2:lockfunction2(args);...",
    )
    db_cmdset_storage = forms.CharField(
        label="CmdSet",
        initial="",
        required=False,
        widget=forms.TextInput(attrs={"size": "78"}),
    )

    # This is not working well because it will not properly allow an empty choice, and will
    # also not work well for comma-separated storage without more work. Notably, it's also
    # a bit hard to visualize.
    # db_cmdset_storage = forms.MultipleChoiceField(
    #     label="CmdSet",
    #     required=False,
    #     choices=adminutils.get_and_load_typeclasses(parent=ObjectDB))

    db_location = forms.ModelChoiceField(
        ObjectDB.objects.all(),
        label="Location",
        required=False,
        widget=ForeignKeyRawIdWidget(
            ObjectDB._meta.get_field("db_location").remote_field, admin.site
        ),
        help_text="The (current) in-game location.<BR>"
        "Usually a Room but can be<BR>"
        "empty for un-puppeted Characters.",
    )
    db_home = forms.ModelChoiceField(
        ObjectDB.objects.all(),
        label="Home",
        required=False,
        widget=ForeignKeyRawIdWidget(
            ObjectDB._meta.get_field("db_location").remote_field, admin.site
        ),
        help_text="Fallback in-game location.<BR>"
        "All objects should usually have<BR>"
        "a home location.",
    )
    db_destination = forms.ModelChoiceField(
        ObjectDB.objects.all(),
        label="Destination",
        required=False,
        widget=ForeignKeyRawIdWidget(
            ObjectDB._meta.get_field("db_destination").remote_field, admin.site
        ),
        help_text="Only used by Exits.",
    )

    def __init__(self, *args, **kwargs):
        """
        Tweak some fields dynamically.

        """
        super().__init__(*args, **kwargs)

        # set default home
        home_id = str(settings.DEFAULT_HOME)
        home_id = home_id[1:] if home_id.startswith("#") else home_id
        default_home = ObjectDB.objects.filter(id=home_id)
        if default_home:
            default_home = default_home[0]
        self.fields["db_home"].initial = default_home
        self.fields["db_location"].initial = default_home

        # better help text for cmdset_storage
        char_cmdset = settings.CMDSET_CHARACTER
        account_cmdset = settings.CMDSET_ACCOUNT
        self.fields["db_cmdset_storage"].help_text = (
            "Path to Command-set path. Most non-character objects don't need a cmdset"
            " and can leave this field blank. Default cmdset-path<BR> for Characters "
            f"is <strong>{char_cmdset}</strong> ."
        )


class ObjectEditForm(ObjectCreateForm):
    """
    Form used for editing. Extends the create one with more fields

    """

    class Meta:
        model = ObjectDB
        fields = "__all__"

    db_account = forms.ModelChoiceField(
        AccountDB.objects.all(),
        label="Puppeting Account",
        required=False,
        widget=ForeignKeyRawIdWidget(
            ObjectDB._meta.get_field("db_account").remote_field, admin.site
        ),
        help_text="An Account puppeting this Object (if any).<BR>Note that when a user logs "
        "off/unpuppets, this<BR>field will be empty again. This is normal.",
    )


@admin.register(ObjectDB)
class ObjectAdmin(admin.ModelAdmin):
    """
    Describes the admin page for Objects.

    """

    inlines = [ObjectTagInline, ObjectAttributeInline]
    list_display = (
        "id",
        "db_key",
        "db_typeclass_path",
        "db_location",
        "db_destination",
        "db_account",
        "db_date_created",
    )
    list_display_links = ("id", "db_key")
    ordering = ["-db_date_created", "-id"]
    search_fields = [
        "=id",
        "^db_key",
        "db_typeclass_path",
        "^db_account__db_key",
        "^db_location__db_key",
    ]
    raw_id_fields = ("db_destination", "db_location", "db_home", "db_account")
    readonly_fields = ("serialized_string", "link_button")

    save_as = True
    save_on_top = True
    list_select_related = True
    view_on_site = False
    list_filter = ("db_typeclass_path",)

    # editing fields setup

    form = ObjectEditForm
    fieldsets = (
        (
            None,
            {
                "fields": (
                    ("db_key", "db_typeclass_path"),
                    ("db_location", "db_home", "db_destination"),
                    ("db_account", "link_button"),
                    "db_cmdset_storage",
                    "db_lock_storage",
                    "serialized_string",
                )
            },
        ),
    )

    add_form = ObjectCreateForm
    add_fieldsets = (
        (
            None,
            {
                "fields": (
                    ("db_key", "db_typeclass_path"),
                    ("db_location", "db_home", "db_destination"),
                    "db_cmdset_storage",
                )
            },
        ),
    )

    def serialized_string(self, obj):
        """
        Get the serialized version of the object.

        """
        from evennia.utils import dbserialize

        return str(dbserialize.pack_dbobj(obj))

    serialized_string.help_text = (
        "Copy & paste this string into an Attribute's `value` field to store this object there."
    )

    def get_fieldsets(self, request, obj=None):
        """
        Return fieldsets.

        Args:
            request (Request): Incoming request.
            obj (Object, optional): Database object.
        """
        if not obj:
            return self.add_fieldsets
        return super().get_fieldsets(request, obj)

    def get_form(self, request, obj=None, **kwargs):
        """
        Use special form during creation.

        Args:
            request (Request): Incoming request.
            obj (Object, optional): Database object.

        """
        help_texts = kwargs.get("help_texts", {})
        help_texts["serialized_string"] = self.serialized_string.help_text
        kwargs["help_texts"] = help_texts

        defaults = {}
        if obj is None:
            defaults.update(
                {"form": self.add_form, "fields": flatten_fieldsets(self.add_fieldsets)}
            )
        defaults.update(kwargs)
        return super().get_form(request, obj, **defaults)

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "account-object-link/<int:pk>",
                self.admin_site.admin_view(self.link_object_to_account),
                name="object-account-link",
            )
        ]
        return custom_urls + urls

    def link_button(self, obj):
        return format_html(
            '<a class="button" href="{}">Link to Account</a>&nbsp;',
            reverse("admin:object-account-link", args=[obj.pk]),
        )

    link_button.short_description = "Create attrs/locks for puppeting"
    link_button.allow_tags = True

    def link_object_to_account(self, request, object_id):
        """
        Link object and account when pressing the button.

        This will:

        - Set account.db._last_puppet to this object
        - Add object to account.characters
        - Change object locks to allow puppeting by account

        """
        obj = self.get_object(request, object_id)
        account = obj.db_account

        if account:
            account.db._last_puppet = obj
            account.characters.add(obj)
            if not obj.access(account, "puppet"):
                lock = obj.locks.get("puppet")
                lock += f" or pid({account.id})"
                obj.locks.add(lock)
            self.message_user(
                request,
                "Did the following (where possible): "
                f"Set Account.db._last_puppet = {obj}, "
                f"Added {obj} to Account.characters list, "
                f"Added 'puppet:pid({account.id})' lock to {obj}.",
            )
        else:
            self.message_user(
                request,
                "Account must be connected for this action "
                "(set Puppeting Account and save this page first).",
                level=messages.ERROR,
            )

        # stay on the same page
        return HttpResponseRedirect(reverse("admin:objects_objectdb_change", args=[obj.pk]))

    def save_model(self, request, obj, form, change):
        """
        Model-save hook.

        Args:
            request (Request): Incoming request.
            obj (Object): Database object.
            form (Form): Form instance.
            change (bool): If this is a change or a new object.

        """
        if not change:
            # adding a new object
            # have to call init with typeclass passed to it
            obj.set_class_from_typeclass(typeclass_path=obj.db_typeclass_path)
            obj.save()
            obj.basetype_setup()
            obj.basetype_posthook_setup()
            obj.at_object_creation()
        else:
            obj.save()
            obj.at_init()

    def response_add(self, request, obj, post_url_continue=None):
        from django.http import HttpResponseRedirect
        from django.urls import reverse

        return HttpResponseRedirect(reverse("admin:objects_objectdb_change", args=[obj.id]))



"""
Unit tests for verb conjugation.

"""

from django.test import TestCase
from parameterized import parameterized

from . import conjugate, pronouns


class TestVerbConjugate(TestCase):
    """
    Test the conjugation.

    """

    @parameterized.expand(
        [
            ("have", "have"),
            ("swim", "swim"),
            ("give", "give"),
            ("given", "give"),
            ("am", "be"),
            ("doing", "do"),
            ("are", "be"),
        ]
    )
    def test_verb_infinitive(self, verb, expected):
        """
        Test the infinite-getter.
        """
        self.assertEqual(expected, conjugate.verb_infinitive(verb))

    @parameterized.expand(
        [
            ("inf", "have", "have"),
            ("inf", "swim", "swim"),
            ("inf", "give", "give"),
            ("inf", "given", "give"),
            ("inf", "am", "be"),
            ("inf", "doing", "do"),
            ("inf", "are", "be"),
            ("2sgpres", "am", "are"),
            ("3sgpres", "am", "is"),
        ]
    )
    def test_verb_conjugate(self, tense, verb, expected):
        """
        Test conjugation for different tenses.

        """
        self.assertEqual(expected, conjugate.verb_conjugate(verb, tense=tense))

    @parameterized.expand(
        [
            ("1st", "have", "have"),
            ("1st", "swim", "swim"),
            ("1st", "give", "give"),
            ("1st", "given", "give"),
            ("1st", "am", "am"),
            ("1st", "doing", "do"),
            ("1st", "are", "am"),
            ("2nd", "were", "are"),
            ("3rd", "am", "is"),
        ]
    )
    def test_verb_present(self, person, verb, expected):
        """
        Test the present.

        """
        self.assertEqual(expected, conjugate.verb_present(verb, person=person))

    @parameterized.expand(
        [
            ("have", "having"),
            ("swim", "swimming"),
            ("give", "giving"),
            ("given", "giving"),
            ("am", "being"),
            ("doing", "doing"),
            ("are", "being"),
        ]
    )
    def test_verb_present_participle(self, verb, expected):
        """
        Test the present_participle

        """
        self.assertEqual(expected, conjugate.verb_present_participle(verb))

    @parameterized.expand(
        [
            ("1st", "have", "had"),
            ("1st", "swim", "swam"),
            ("1st", "give", "gave"),
            ("1st", "given", "gave"),
            ("1st", "am", "was"),
            ("1st", "doing", "did"),
            ("1st", "are", "was"),
            ("2nd", "were", "were"),
        ]
    )
    def test_verb_past(self, person, verb, expected):
        """
        Test the past getter.

        """
        self.assertEqual(expected, conjugate.verb_past(verb, person=person))

    @parameterized.expand(
        [
            ("have", "had"),
            ("swim", "swum"),
            ("give", "given"),
            ("given", "given"),
            ("am", "been"),
            ("doing", "done"),
            ("are", "been"),
        ]
    )
    def test_verb_past_participle(self, verb, expected):
        """
        Test the past participle.

        """
        self.assertEqual(expected, conjugate.verb_past_participle(verb))

    def test_verb_get_all_tenses(self):
        """
        Test getting all tenses.

        """
        self.assertEqual(list(conjugate.verb_tenses_keys.keys()), conjugate.verb_all_tenses())

    @parameterized.expand(
        [
            ("have", "infinitive"),
            ("swim", "infinitive"),
            ("give", "infinitive"),
            ("given", "past participle"),
            ("am", "1st singular present"),
            ("doing", "present participle"),
            ("are", "2nd singular present"),
        ]
    )
    def test_verb_tense(self, verb, expected):
        """
        Test the tense retriever.

        """
        self.assertEqual(expected, conjugate.verb_tense(verb))

    @parameterized.expand(
        [
            ("inf", "have", True),
            ("inf", "swim", True),
            ("inf", "give", True),
            ("inf", "given", False),
            ("inf", "am", False),
            ("inf", "doing", False),
            ("inf", "are", False),
        ]
    )
    def test_verb_is_tense(self, tense, verb, expected):
        """
        Test the tense-checker

        """
        self.assertEqual(expected, conjugate.verb_is_tense(verb, tense))

    @parameterized.expand(
        [
            ("1st", "have", False),
            ("1st", "swim", False),
            ("1st", "give", False),
            ("1st", "given", False),
            ("1st", "am", True),
            ("1st", "doing", False),
            ("1st", "are", False),
            ("1st", "had", False),
        ]
    )
    def test_verb_is_present(self, person, verb, expected):
        """
        Test the tense-checker

        """
        self.assertEqual(expected, conjugate.verb_is_present(verb, person=person))

    @parameterized.expand(
        [
            ("have", False),
            ("swim", False),
            ("give", False),
            ("given", False),
            ("am", False),
            ("doing", True),
            ("are", False),
        ]
    )
    def test_verb_is_present_participle(self, verb, expected):
        """
        Test the tense-checker

        """
        self.assertEqual(expected, conjugate.verb_is_present_participle(verb))

    @parameterized.expand(
        [
            ("1st", "have", False),
            ("1st", "swim", False),
            ("1st", "give", False),
            ("1st", "given", False),
            ("1st", "am", False),
            ("1st", "doing", False),
            ("1st", "are", False),
            ("2nd", "were", True),
        ]
    )
    def test_verb_is_past(self, person, verb, expected):
        """
        Test the tense-checker

        """
        self.assertEqual(expected, conjugate.verb_is_past(verb, person=person))

    @parameterized.expand(
        [
            ("have", False),
            ("swimming", False),
            ("give", False),
            ("given", True),
            ("am", False),
            ("doing", False),
            ("are", False),
            ("had", False),
        ]
    )
    def test_verb_is_past_participle(self, verb, expected):
        """
        Test the tense-checker

        """
        self.assertEqual(expected, conjugate.verb_is_past_participle(verb))

    @parameterized.expand(
        [
            ("have", ("have", "has")),
            ("swimming", ("swimming", "swimming")),
            ("give", ("give", "gives")),
            ("given", ("given", "given")),
            ("am", ("are", "is")),
            ("doing", ("doing", "doing")),
            ("are", ("are", "is")),
            ("had", ("had", "had")),
            ("grin", ("grin", "grins")),
            ("smile", ("smile", "smiles")),
            ("vex", ("vex", "vexes")),
            ("thrust", ("thrust", "thrusts")),
        ]
    )
    def test_verb_actor_stance_components(self, verb, expected):
        """
        Test the tense-checker

        """
        self.assertEqual(expected, conjugate.verb_actor_stance_components(verb))


class TestPronounMapping(TestCase):
    """
    Test pronoun viewpoint mapping
    """

    @parameterized.expand(
        [
            ("you", "you", "it"),  # default 3rd is "neutral"
            ("I", "I", "it"),
            ("Me", "Me", "It"),
            ("ours", "ours", "theirs"),
            ("yourself", "yourself", "itself"),
            ("yourselves", "yourselves", "themselves"),
            ("he", "you", "he"),  # assume 2nd person
            ("her", "you", "her"),
            ("their", "your", "their"),
            ("itself", "yourself", "itself"),
            ("herself", "yourself", "herself"),
            ("themselves", "yourselves", "themselves"),
        ]
    )
    def test_default_mapping(self, pronoun, expected_1st_or_2nd_person, expected_3rd_person):
        """
        Test the pronoun mapper.

        """
        received_1st_or_2nd_person, received_3rd_person = pronouns.pronoun_to_viewpoints(pronoun)

        self.assertEqual(expected_1st_or_2nd_person, received_1st_or_2nd_person)
        self.assertEqual(expected_3rd_person, received_3rd_person)

    @parameterized.expand(
        [
            ("you", "m", "you", "he"),
            ("you", "f op", "you", "her"),
            ("you", "p op", "you", "them"),
            ("I", "m", "I", "he"),
            ("Me", "n", "Me", "It"),
            ("your", "p", "your", "their"),
            ("yourself", "m", "yourself", "himself"),
            ("yourself", "f", "yourself", "herself"),
            ("yourselves", "", "yourselves", "themselves"),
            ("he", "1", "I", "he"),
            ("he", "1 p", "we", "he"),  # royal we
            ("we", "m", "we", "he"),  # royal we, other way
            ("her", "p", "you", "her"),
            ("her", "pa", "your", "her"),
            ("their", "ma", "your", "their"),
        ]
    )
    def test_mapping_with_options(
        self, pronoun, options, expected_1st_or_2nd_person, expected_3rd_person
    ):
        """
        Test the pronoun mapper.

        """
        received_1st_or_2nd_person, received_3rd_person = pronouns.pronoun_to_viewpoints(
            pronoun, options
        )
        self.assertEqual(expected_1st_or_2nd_person, received_1st_or_2nd_person)
        self.assertEqual(expected_3rd_person, received_3rd_person)

    @parameterized.expand(
        [
            ("you", "p", "you", "they"),
            ("I", "p", "I", "they"),
            ("Me", "p", "Me", "Them"),
            ("your", "p", "your", "their"),
            ("they", "1 p", "we", "they"),
            ("they", "", "you", "they"),
            ("yourself", "p", "yourself", "themselves"),
            ("myself", "p", "myself", "themselves"),
        ]
    )
    def test_colloquial_plurals(
        self, pronoun, options, expected_1st_or_2nd_person, expected_3rd_person
    ):
        """
        The use of this module by the funcparser expects a default person-pronoun
        of the neutral "they", which is categorized here by the plural.

        """
        received_1st_or_2nd_person, received_3rd_person = pronouns.pronoun_to_viewpoints(
            pronoun, options
        )

        self.assertEqual(expected_1st_or_2nd_person, received_1st_or_2nd_person)
        self.assertEqual(expected_3rd_person, received_3rd_person)



"""

Telnet OOB (Out of band communication)

OOB protocols allow for asynchronous communication between Evennia and
compliant telnet clients. The "text" type of send command will always
be sent "in-band", appearing in the client's main text output. OOB
commands, by contrast, can have many forms and it is up to the client
how and if they are handled.  Examples of OOB instructions could be to
instruct the client to play sounds or to update a graphical health
bar.

Note that in Evennia's Web client, all send commands are "OOB
commands", (including the "text" one), there is no equivalence to
MSDP/GMCP for the webclient since it doesn't need it.

This implements the following telnet OOB communication protocols:

- MSDP (Mud Server Data Protocol), as per http://tintin.sourceforge.net/msdp/
- GMCP (Generic Mud Communication Protocol) as per
  http://www.ironrealms.com/rapture/manual/files/FeatGMCP-txt.html#Generic_MUD_Communication_Protocol%28GMCP%29

----

"""
import json
import re

# General Telnet
from twisted.conch.telnet import IAC, SB, SE

from evennia.utils.utils import is_iter

# MSDP-relevant telnet cmd/opt-codes
MSDP = bytes([69])
MSDP_VAR = bytes([1])
MSDP_VAL = bytes([2])
MSDP_TABLE_OPEN = bytes([3])
MSDP_TABLE_CLOSE = bytes([4])

MSDP_ARRAY_OPEN = bytes([5])
MSDP_ARRAY_CLOSE = bytes([6])

# GMCP
GMCP = bytes([201])


# pre-compiled regexes
# returns 2-tuple
msdp_regex_table = re.compile(
    rb"%s\s*(\w*?)\s*%s\s*%s(.*?)%s" % (MSDP_VAR, MSDP_VAL, MSDP_TABLE_OPEN, MSDP_TABLE_CLOSE)
)
# returns 2-tuple
msdp_regex_array = re.compile(
    rb"%s\s*(\w*?)\s*%s\s*%s(.*?)%s" % (MSDP_VAR, MSDP_VAL, MSDP_ARRAY_OPEN, MSDP_ARRAY_CLOSE)
)
msdp_regex_var = re.compile(rb"%s" % MSDP_VAR)
msdp_regex_val = re.compile(rb"%s" % MSDP_VAL)

EVENNIA_TO_GMCP = {
    "client_options": "Core.Supports.Get",
    "get_inputfuncs": "Core.Commands.Get",
    "get_value": "Char.Value.Get",
    "repeat": "Char.Repeat.Update",
    "monitor": "Char.Monitor.Update",
}


# MSDP/GMCP communication handler


class TelnetOOB:
    """
    Implements the MSDP and GMCP protocols.
    """

    def __init__(self, protocol):
        """
        Initiates by storing the protocol on itself and trying to
        determine if the client supports MSDP.

        Args:
            protocol (Protocol): The active protocol.

        """
        self.protocol = protocol
        self.protocol.protocol_flags["OOB"] = False
        self.MSDP = False
        self.GMCP = False
        # ask for the available protocols and assign decoders
        # (note that handshake_done() will be called twice!)
        self.protocol.negotiationMap[MSDP] = self.decode_msdp
        self.protocol.negotiationMap[GMCP] = self.decode_gmcp
        self.protocol.will(MSDP).addCallbacks(self.do_msdp, self.no_msdp)
        self.protocol.will(GMCP).addCallbacks(self.do_gmcp, self.no_gmcp)
        self.oob_reported = {}

    def no_msdp(self, option):
        """
        Client reports No msdp supported or wanted.

        Args:
            option (Option): Not used.

        """
        # no msdp, check GMCP
        self.protocol.handshake_done()

    def do_msdp(self, option):
        """
        Client reports that it supports msdp.

        Args:
            option (Option): Not used.

        """
        self.MSDP = True
        self.protocol.protocol_flags["OOB"] = True
        self.protocol.handshake_done()

    def no_gmcp(self, option):
        """
        If this is reached, it means neither MSDP nor GMCP is
        supported.

        Args:
            option (Option): Not used.

        """
        self.protocol.handshake_done()

    def do_gmcp(self, option):
        """
        Called when client confirms that it can do MSDP or GMCP.

        Args:
            option (Option): Not used.

        """
        self.GMCP = True
        self.protocol.protocol_flags["OOB"] = True
        self.protocol.handshake_done()

    # encoders

    def encode_msdp(self, cmdname, *args, **kwargs):
        """
        Encode into a valid MSDP command.

        Args:
            cmdname (str): Name of send instruction.
            args, kwargs (any): Arguments to OOB command.

        Notes:
            The output of this encoding will be
            MSDP structures on these forms:
            ::

                [cmdname, [], {}]           -> VAR cmdname VAL ""
                [cmdname, [arg], {}]        -> VAR cmdname VAL arg
                [cmdname, [args],{}]        -> VAR cmdname VAL ARRAYOPEN VAL arg VAL arg ... ARRAYCLOSE
                [cmdname, [], {kwargs}]     -> VAR cmdname VAL TABLEOPEN VAR key VAL val ... TABLECLOSE
                [cmdname, [args], {kwargs}] -> VAR cmdname VAL ARRAYOPEN VAL arg VAL arg ... ARRAYCLOSE
                                               VAR cmdname VAL TABLEOPEN VAR key VAL val ... TABLECLOSE

            Further nesting is not supported, so if an array argument
            consists of an array (for example), that array will be
            json-converted to a string.

        """
        msdp_cmdname = "{msdp_var}{msdp_cmdname}{msdp_val}".format(
            msdp_var=MSDP_VAR.decode(), msdp_cmdname=cmdname, msdp_val=MSDP_VAL.decode()
        )

        if not (args or kwargs):
            return msdp_cmdname.encode()

        # print("encode_msdp in:", cmdname, args, kwargs)  # DEBUG

        msdp_args = ""
        if args:
            msdp_args = msdp_cmdname
            if len(args) == 1:
                msdp_args += args[0]
            else:
                msdp_args += (
                    "{msdp_array_open}"
                    "{msdp_args}"
                    "{msdp_array_close}".format(
                        msdp_array_open=MSDP_ARRAY_OPEN.decode(),
                        msdp_array_close=MSDP_ARRAY_CLOSE.decode(),
                        msdp_args="".join("%s%s" % (MSDP_VAL.decode(), val) for val in args),
                    )
                )

        msdp_kwargs = ""
        if kwargs:
            msdp_kwargs = msdp_cmdname
            msdp_kwargs += (
                "{msdp_table_open}"
                "{msdp_kwargs}"
                "{msdp_table_close}".format(
                    msdp_table_open=MSDP_TABLE_OPEN.decode(),
                    msdp_table_close=MSDP_TABLE_CLOSE.decode(),
                    msdp_kwargs="".join(
                        "%s%s%s%s" % (MSDP_VAR.decode(), key, MSDP_VAL.decode(), val)
                        for key, val in kwargs.items()
                    ),
                )
            )

        msdp_string = msdp_args + msdp_kwargs

        # print("msdp_string:", msdp_string)  # DEBUG
        return msdp_string.encode()

    def encode_gmcp(self, cmdname, *args, **kwargs):
        """
        Encode into GMCP messages.

        Args:
            cmdname (str): GMCP OOB command name.
            args, kwargs (any): Arguments to OOB command.

        Notes:
            GMCP messages will be outgoing on the following
            form (the non-JSON cmdname at the start is what
            IRE games use, supposedly, and what clients appear
            to have adopted). A cmdname without Package will end
            up in the Core package, while Core package names will
            be stripped on the Evennia side.
            ::

                [cmd_name, [], {}]          -> Cmd.Name
                [cmd_name, [arg], {}]       -> Cmd.Name arg
                [cmd_name, [args],{}]       -> Cmd.Name [args]
                [cmd_name, [], {kwargs}]    -> Cmd.Name {kwargs}
                [cmdname, [args, {kwargs}]  -> Core.Cmdname [[args],{kwargs}]

            For more flexibility with certain clients, if `cmd_name` is capitalized,
            Evennia will leave its current capitalization (So CMD_nAmE would be sent
            as CMD.nAmE but cMD_Name would be Cmd.Name)

        Notes:
            There are also a few default mappings between evennia outputcmds and GMCP:
            ::

                client_options -> Core.Supports.Get
                get_inputfuncs -> Core.Commands.Get
                get_value      -> Char.Value.Get
                repeat         -> Char.Repeat.Update
                monitor        -> Char.Monitor.Update

        """

        if cmdname in EVENNIA_TO_GMCP:
            gmcp_cmdname = EVENNIA_TO_GMCP[cmdname]
        elif "_" in cmdname:
            if cmdname.istitle():
                # leave without capitalization
                gmcp_cmdname = ".".join(word for word in cmdname.split("_"))
            else:
                gmcp_cmdname = ".".join(word.capitalize() for word in cmdname.split("_"))
        else:
            gmcp_cmdname = "Core.%s" % (cmdname if cmdname.istitle() else cmdname.capitalize())

        if not (args or kwargs):
            gmcp_string = gmcp_cmdname
        elif args:
            if len(args) == 1:
                args = args[0]
            if kwargs:
                gmcp_string = "%s %s" % (gmcp_cmdname, json.dumps([args, kwargs]))
            else:
                gmcp_string = "%s %s" % (gmcp_cmdname, json.dumps(args))
        else:  # only kwargs
            gmcp_string = "%s %s" % (gmcp_cmdname, json.dumps(kwargs))

        # print("gmcp string", gmcp_string)  # DEBUG
        return gmcp_string.encode()

    def decode_msdp(self, data):
        """
        Decodes incoming MSDP data.

        Args:
            data (str or list): MSDP data.

        Notes:
            Clients should always send MSDP data on
            one of the following forms:
            ::

                cmdname ''          -> [cmdname, [], {}]
                cmdname val         -> [cmdname, [val], {}]
                cmdname array       -> [cmdname, [array], {}]
                cmdname table       -> [cmdname, [], {table}]
                cmdname array cmdname table -> [cmdname, [array], {table}]

            Observe that all MSDP_VARS are used to identify cmdnames,
            so if there are multiple arrays with the same cmdname
            given, they will be merged into one argument array, same
            for tables. Different MSDP_VARS (outside tables) will be
            identified as separate cmdnames.

        """
        if isinstance(data, list):
            data = b"".join(data)

        # print("decode_msdp in:", data)  # DEBUG

        tables = {}
        arrays = {}
        variables = {}

        # decode tables
        for key, table in msdp_regex_table.findall(data):
            key = key.decode()
            tables[key] = {} if key not in tables else tables[key]
            for varval in msdp_regex_var.split(table)[1:]:
                var, val = msdp_regex_val.split(varval, 1)
                var, val = var.decode(), val.decode()
                if var:
                    tables[key][var] = val

        # decode arrays from all that was not a table
        data_no_tables = msdp_regex_table.sub(b"", data)
        for key, array in msdp_regex_array.findall(data_no_tables):
            key = key.decode()
            arrays[key] = [] if key not in arrays else arrays[key]
            parts = msdp_regex_val.split(array)
            parts = [part.decode() for part in parts]
            if len(parts) == 2:
                arrays[key].append(parts[1])
            elif len(parts) > 1:
                arrays[key].extend(parts[1:])

        # decode remainders from all that were not tables or arrays
        data_no_tables_or_arrays = msdp_regex_array.sub(b"", data_no_tables)
        for varval in msdp_regex_var.split(data_no_tables_or_arrays):
            # get remaining varvals after cleaning away tables/arrays. If mathcing
            # an existing key in arrays, it will be added as an argument to that command,
            # otherwise it will be treated as a command without argument.
            parts = msdp_regex_val.split(varval)
            parts = [part.decode() for part in parts]
            if len(parts) == 2:
                variables[parts[0]] = parts[1]
            elif len(parts) > 1:
                variables[parts[0]] = parts[1:]

        cmds = {}
        # merge matching table/array/variables together
        for key, table in tables.items():
            args, kwargs = [], table
            if key in arrays:
                args.extend(arrays.pop(key))
            if key in variables:
                args.append(variables.pop(key))
            cmds[key] = [args, kwargs]

        for key, arr in arrays.items():
            args, kwargs = arr, {}
            if key in variables:
                args.append(variables.pop(key))
            cmds[key] = [args, kwargs]

        for key, var in variables.items():
            cmds[key] = [[var], {}]

        # remap the 'generic msdp commands' to avoid colliding with builtins etc
        # by prepending "msdp_"
        lower_case = {key.lower(): key for key in cmds}
        for remap in ("list", "report", "reset", "send", "unreport"):
            if remap in lower_case:
                cmds["msdp_{}".format(remap)] = cmds.pop(lower_case[remap])

        # print("msdp data in:", cmds)  # DEBUG
        self.protocol.data_in(**cmds)

    def decode_gmcp(self, data):
        """
        Decodes incoming GMCP data on the form 'varname <structure>'.

        Args:
            data (str or list): GMCP data.

        Notes:
            Clients send data on the form "Module.Submodule.Cmdname <structure>".
            We assume the structure is valid JSON.

            The following is parsed into Evennia's formal structure:
            ::

                Core.Name                         -> [name, [], {}]
                Core.Name string                  -> [name, [string], {}]
                Core.Name [arg, arg,...]          -> [name, [args], {}]
                Core.Name {key:arg, key:arg, ...} -> [name, [], {kwargs}]
                Core.Name [[args], {kwargs}]      -> [name, [args], {kwargs}]

        """
        if isinstance(data, list):
            data = b"".join(data)

        # print("decode_gmcp in:", data)  # DEBUG
        if data:
            try:
                cmdname, structure = data.split(None, 1)
            except ValueError:
                cmdname, structure = data, b""
            cmdname = cmdname.replace(b".", b"_")
            try:
                structure = json.loads(structure)
            except ValueError:
                # maybe the structure is not json-serialized at all
                pass
            args, kwargs = [], {}
            if is_iter(structure):
                if isinstance(structure, dict):
                    kwargs = {key: value for key, value in structure.items() if key}
                else:
                    args = list(structure)
            else:
                args = (structure,)
            if cmdname.lower().startswith(b"core_"):
                # if Core.cmdname, then use cmdname
                cmdname = cmdname[5:]
            self.protocol.data_in(**{cmdname.lower().decode(): [args, kwargs]})

    # access methods

    def data_out(self, cmdname, *args, **kwargs):
        """
        Return a MSDP- or GMCP-valid subnegotiation across the protocol.

        Args:
            cmdname (str): OOB-command name.
            args, kwargs (any): Arguments to OOB command.

        """
        kwargs.pop("options", None)

        if self.MSDP:
            encoded_oob = self.encode_msdp(cmdname, *args, **kwargs)
            self.protocol._write(IAC + SB + MSDP + encoded_oob + IAC + SE)

        if self.GMCP:
            encoded_oob = self.encode_gmcp(cmdname, *args, **kwargs)
            self.protocol._write(IAC + SB + GMCP + encoded_oob + IAC + SE)
